import numpy as np
import qetpy as qp
import itertools


__all__ = [
    'PileupOF',
]


class PileupOF(object):
    """
    Class for efficient calculation of the two-pulse Pileup Optimum
    Filter (OF).

    Attributes
    ----------
    pileup_res : ndarray, NoneType
        Calculated by the `run` class method, this contains the
        amplitudes, time-shift values, and chi-square from the
        pileup optimum filter within the range of time values
        where two pulses would have a large amount of interference,
        nonnegligibly reducing the sensitivity of the iterative pileup
        optimum filter. Set to None when initialized or when the signal
        is updated via the `update_signal` class method.

    Notes
    -----
    The parameters in both `pileup_res` are:

        (amplitude 1, time offset 1,
         amplitude 2, time offset 2, chi-square)

    """

    def __init__(self, signal, template, psd, fs, error_cutoff=0.1,
                 ac_coupled=True, integralnorm=False):
        """
        Initalization of the PileupOF class.

        Parameters
        ----------
        signal : ndarray
            The signal that we want to apply the pileup optimum filter
            to (units should be Amps).
        template : ndarray
            The pulse template to be used for the pileup optimum filter
            (should be normalized to a max height of 1 beforehand).
        psd : ndarray
            The two-sided psd that will be used to describe the noise
            in the signal (in Amps^2/Hz)
        fs : ndarray
            The sample rate of the data being taken (in Hz).
        error_cutoff : float
            The cutoff on the error in the inversion of the pileup OF
            matrix to determine the time range to consider pileup.
            Setting to small values corresponds to a larger time
            range. Default is 0.1.
        ac_coupled : bool, optional
            If True, the zero frequency bin of the psd
            will be ignored (i.e. set to infinity) when calculating
            the optimum amplitude. If False, then the zero frequency
            bin is kept. Default is True.
        integralnorm : bool, optional
            If set to True, then the template will be normalized to
            have an integral of 1, and any optimum filters will instead
            return the optimum integral in units of Coulombs. If set to
            False, then the usual optimum filter amplitudes will be
            returned (in units of Amps). Default is False.

        """

        self._psd = np.zeros(len(psd))
        self._psd[:] = psd
        self._psd0 = psd[0]

        if ac_coupled:
            self._psd[0] = np.inf

        self._nbins = signal.shape[-1]
        self._fs = fs
        self._df = self._fs / self._nbins

        self._s = np.fft.fft(template) / self._nbins / self._df

        if integralnorm:
            self._s /= self._s[0]

        self._phi = self._s.conjugate() / self._psd
        self._norm = np.real(np.dot(self._phi, self._s)) * self._df

        self._pmatrix_off = np.real(
            np.fft.ifft(self._s * self._phi) / self._norm * self._fs
        )[:self._nbins//2]

        self._freqs = np.fft.fftfreq(self._nbins, d=1 / self._fs)

        self._tcutoff = self._determine_tcutoff(error_cutoff)

        self._createpmatrices()

        self._OF = qp.OptimumFilter(
            signal,
            template,
            psd,
            fs,
            coupling="AC" if ac_coupled else "DC",
            integralnorm=integralnorm,
        )

        self._update_signal(signal)


    def _update_signal(self, signal):
        """
        Hidden function for recalculating parameters that depend on the
        signal.

        """

        self.pileup_res = None
        self._v = np.fft.fft(signal, axis=-1) / self._nbins / self._df
        self._qn = np.real(
            np.fft.ifft(self._v * self._phi) / self._norm * self._fs
        )

        amp_start, t0_start, chi2_start = self._OF.ofamp_withdelay()

        self._time_array = np.arange(
            -(int(2 * self._tcutoff * self._fs)//2) + int(t0_start * self._fs),
            int(2 * self._tcutoff * self._fs)//2 + int(
                2 * self._tcutoff * self._fs
            )%2 + int(t0_start * self._fs),
        ) / self._fs

        self._exp_times = np.roll(
            self._time_array,
            -(int(2 * self._tcutoff * self._fs)//2) + int(t0_start * self._fs),
        )
        self._exp_array = np.exp(
            -2.0j * np.pi * self._exp_times[:, None] * self._freqs[None, :]
        )


    def update_signal(self, signal):
        """
        Function for updating all parameters that are signal-dependent,
        so that the pileup optimum filter can be run on new data
        without recalculating parameters unnecessarily.

        Parameters
        ----------
        signal : ndarray
            The signal that we want to apply the pileup optimum filter
            to (units should be Amps).

        Returns
        -------
        None

        """

        self._OF.update_signal(signal)
        self._update_signal(signal)


    def _determine_tcutoff(self, cutoff):
        """
        Hidden function to determine the cutoff on the time difference
        between pulses where, after this point, the iterative pileup
        optimum filter is "good enough".

        """

        tcutoff = np.argmax(self._pmatrix_off < cutoff) / self._fs

        return tcutoff


    def _createpmatrices(self):
        """
        Hidden function to facilitate the precalculation of the
        matrices used to determine the pileup amplitudes.

        """

        npulses = 2

        end_ind = int(2 * self._tcutoff * self._fs)
        self._p = np.zeros((end_ind, npulses, npulses))
        np.einsum('jii->ji', self._p)[:] = 1
        self._p[:, 0, 1] = self._p[:, 1, 0] = self._pmatrix_off[:end_ind]

        self._p_inv = np.linalg.pinv(self._p)
        self._p_inv[0] = np.array([[1, 0], [0, 0]])


    def _get_amps(self, t0s):
        """
        Hidden function to calculate the amplitudes that correspond to
        the inputted time offsets.

        """

        p_ind = int(np.abs(t0s[0] - t0s[1]) * self._fs)
        pmatrix_inv = self._p_inv[p_ind]
        qvec = np.array(
            [self._qn[int(t * self._fs)] for t in t0s]
        )

        return pmatrix_inv @ qvec


    def _chi2(self, amps, t0s):
        """
        Hidden function to calculate the chi-square of the inputted
        amplitude and time offsets.

        """

        numer = self._v - self._s * sum(
            a * self._exp_array[
                np.argmin(np.abs(self._exp_times - t))
            ] for a, t in zip(amps, t0s)
        )

        return np.real(np.dot(numer.conjugate() / self._psd, numer) * self._df)


    def run(self):
        """
        Runs the pileup optimum filter algorithm for 2 pulses.

        Parameters
        ----------
        None

        Returns
        -------
        res : ndarray
            The results of the pileup optimum filter algorithm. The
            parameters returned are (amplitude 1, time offset 1,
            amplitude 2, time offset 2, chi-square).

        """

        npulses = 2

        combs = list(itertools.combinations(self._time_array, npulses))

        pileup_results = np.zeros((len(combs), 2 * npulses + 1))

        for ii, t0s in enumerate(combs):
            amps = self._get_amps(t0s)
            chi2 = self._chi2(amps, t0s)
            results = [None] * npulses * 2
            results[::2] = amps
            results[1::2] = t0s
            results.append(chi2)
            pileup_results[ii] = np.array(results)

        self.pileup_res = pileup_results[np.argmin(pileup_results[:, -1])]

        return self.pileup_res



import numpy as np
import matplotlib.pyplot as plt
from qetpy.utils import lowpassfilter


__all__ = [
    "plotnsmb",
]


def plotnsmb(pulset, fs, tdelmin, amin, sbTemplatef, nS, nB, nt, psddnu,
             lpFiltFreq=None, lgcsaveplots=False, xlim=None, figPrefix='testFit',
             background_templates_shifts=None, savepath='./'):
    """
    Diagnostic plotting function for the OF nSmB code.

    Parameters
    ----------
    pulset : tuple
        Data to be plotted
        Dimensions: 1 X (time bins)
    omega : tuple
        Angular frequency of the pulse in the frequency domain
        Dimensions: (time bins) X ()
    fs : tuple
        Sampling frequency in Hz
    tdelmin : tuple
        The best fit time delay of the signal, offset from the original signal template
    amin : tuple
        The best fit amplitude of the signal
        Dimensions: (nS + nB) X 1
    sbTemplatef : ndarray
        The frequency domain signal and background templates
        Dimensions: (nS + nB) X (time bins)
    nt : tuple
        The number of time domain points of the signal and template
    lpFiltFreq : tuple, optional
        The frequency of the LF filter to pulset (default None)
    lgcsaveplots : bool, int, optional
        Flag for whether or not to save plots (default False)
        If not False, should be int to append to saved fig filename
    figPrefix : str, optional
        The saved fig filename prefix (whatever text before the number)
    background_templates_shifts : ndarray, optional
        The indices at which the background templates start
        Dimensions: m X ()
    savepath : str, optional
        Path to save directory. Default is current working directory.

    """

    nSB = nS + nB

    if (lpFiltFreq!=None):
        pulseFilt = lowpassfilter(pulset, lpFiltFreq, fs)

    # === DAQ Setup ===
    dt = float(1) / fs
    dnu = float(1) / (nt * dt)
    nu = np.arange(0, float(nt)) * dnu
    lgc = nu > nt * dnu / 2
    nu[lgc] = nu[lgc] - nt * dnu
    omega = (2 * np.pi) * nu

    # create a phase shift matrix
    # The signal gets phase shifted by tdelmin
    # The background templates have no phase shift
    phase = np.exp(-1j * omega * tdelmin)
    phaseAr = np.ones((nS, 1)) @ phase[None,:]
    phaseMat= np.concatenate((phaseAr, np.ones((nB, nt))), axis=0)
    ampMat = amin @ np.ones((1, nt))
    fitf = ampMat * sbTemplatef * phaseMat
    fittotf = np.sum(fitf, axis=0, keepdims=True)

    # get baseline of pulset
    pulsetBL = np.mean(pulset[0,0:300])

    # now invert
    fitt = np.real(np.fft.ifft(fitf, axis=1) * nt)
    fittott = np.real(np.fft.ifft(fittotf, axis=1) * nt);

    # make residual 
    residT = pulset - fittott
    # check the chi2
    residTf = np.fft.fft(residT, axis=1) / nt
    chi2T = np.real(np.sum(np.conj(residTf.T) / psddnu.T * residTf.T, 0))


    chi2TFloat = float(chi2T)

    # ===Time Domain ==================================================
    bins = np.arange(nt)
    bins = bins[None,:]
    timeP = bins * dt
    plt.figure(figsize=(12, 7));
    if (lpFiltFreq):
        plt.plot(bins.T, pulseFilt.T, '-k', label='Data (LP filtered)')
        plt.plot(bins.T, pulset.T, '-b', alpha=0.4, linewidth=0.5, label='Data')
    else:
        plt.plot(bins.T, pulset.T, '-k', label='Data')
    plt.plot(bins.T, fittott.T, '-g', label=f'Total Fit. $\chi^2$={chi2TFloat:.1f}', linewidth=5)    
    # plot the background template best fit
    # note:
    #    1) the loop starts at nS and only goes to nSB - 2
    #       thereby leaving out the DC and slope component
    #    2) the DC and slope components are added to the
    #       pulse background templates
    for jSB in range(nS, nSB - 2):
        if(jSB==nS):
            plt.plot(
                bins.T,
                fitt[jSB, :, None] + fitt[-1, :, None] + fitt[-2, :, None],
                '-',
                c='r',
                label='Background Fit',
            )
        else:
            plt.plot(
                bins.T,
                fitt[jSB, :, None] + fitt[-1, :, None] + fitt[-2, :, None],
                '-',
                c='r',
            )
    # plot the signal template fit if the amplitude
    # is larger than a small threshold
    if (np.abs(amin[0]) > 1e-20):
        plt.plot(
            bins.T,
            fitt[0, :, None]  + fitt[-1, :, None] + fitt[-2, :, None],
            '-c',
            label='Signal Fit',
        )

    if background_templates_shifts is not None:
            for ii in range(nB):
                if (ii==0):
                    plt.axvline(
                        x=background_templates_shifts[ii],
                        linestyle='--',
                        color='m',
                        linewidth=1,
                        label='Bckg. Start Times',
                    )
                else:
                    plt.axvline(
                        x=background_templates_shifts[ii],
                        linestyle='--',
                        color='m',
                        linewidth=1,
                    )

    plt.xlabel(f'bins (1 bin = {(dt * 1e6):.2f} $\mu$s)');
    plt.ylabel('Amps');

    if xlim is not None: 
        plt.xlim(xlim)
    plt.legend()
    plt.grid()
    plt.ticklabel_format(style='sci', axis='both', scilimits=(0, 0))

    if lgcsaveplots:
        plt.savefig(savepath + figPrefix + str(lgcsaveplots) + '.png', bbox_inches='tight')



import numpy as np
import matplotlib.pyplot as plt


__all__ = [
    "plot_iv",
    "plot_rv",
    "plot_pv",
    "plot_all_curves",
]


def plot_iv(IVobject, temps="all", chans="all", showfit=True, lgcsave=False, savepath="", savename=""):
    """
    Function to plot the IV curves for the data in an IV object.

    Parameters
    ----------
    IVobject : class
        The IV class object that the data is stored in.
    temps : string, array_like, int, optional
        Which bath temperatures to plot. Setting to "all" plots all of them. Can also set
        to a subset of bath temperatures, or just one
    chans : string, array_like, int, optional
        Which bath temperatures to plot. Setting to "all" plots all of them. Can also set
        to a subset of bath temperatures, or just one
    showfit : boolean, optional
        Boolean flag to also plot the linear fit to the normal data
    lgcsave : boolean, optional
        Boolean flag to save the plot
    savepath : string, optional
        Path to save the plot to, saves it to the current directory by default
    savename : string, optional
        Name to append to the plot file name, if saving

    """

    ntemps, nch, niters = IVobject.dites.shape
    chan_names = IVobject.chan_names

    if temps == "all":
        trange = range(ntemps)
    elif np.isscalar(temps):
        trange = np.array([temps])
    else:
        trange = np.array(temps)

    if chans == "all":
        chrange = range(nch)
    elif np.isscalar(chans):
        chrange = np.array([chans])
    else:
        chrange = np.array(chans)

    ch_colors = plt.cm.viridis(np.linspace(0, 1, num=len(trange) * len(chrange)))

    fig, ax = plt.subplots(figsize=(10, 6))

    for it, t in enumerate(trange):
        for ich, ch in enumerate(chrange):
            if ntemps > 1:
                label_str="Temp {}, Channel {}".format(t, chan_names[ch])
            else:
                label_str="Channel {}".format(chan_names[ch])
            ax.scatter(
                IVobject.vb[t, ch] * 1e6,
                IVobject.ites[t, ch] * 1e6,
                label=label_str,
                color=ch_colors[it * len(chrange) + ich],
                s=10.0,
            )
            ax.plot(
                IVobject.vb[t, ch] * 1e6,
                IVobject.ites[t, ch] * 1e6,
                color=ch_colors[it * len(chrange) + ich],
                alpha=0.5,
            )
            ax.errorbar(
                IVobject.vb[t, ch] * 1e6,
                IVobject.ites[t, ch] * 1e6,
                yerr=IVobject.ites_err[t, ch] * 1e6,
                linestyle='None',
                color='k',
            )
            if showfit:
                maxind = np.argmax(abs(IVobject.vb[t, ch]))

                if IVobject.vb[t, ch, maxind] > 0:
                    vbfit = np.linspace(0, max(IVobject.vb[t,ch]*1e6), num=10)
                else:
                    vbfit = np.linspace(min(IVobject.vb[t,ch]*1e6), 0, num=10)

                ax.plot(
                    vbfit,
                    1.0 / (IVobject.rfit[t, ch]) * vbfit,
                    color=ch_colors[it * len(chrange) + ich],
                    alpha=0.1,
                )

    ax.legend(loc='best')
    ax.set_xlabel(r'Bias Voltage [$\mu V$]')
    ax.set_ylabel(r'Current through TES [$\mu A$]')
    ax.grid(linestyle='dotted')
    ax.tick_params(which='both', direction='in', right=True, top=True)
    ax.set_title("$I_0$ vs. $V_b$")

    if lgcsave:
        fig.savefig(savepath + "iv_curve_{}.png".format(savename))
        plt.close(fig)
    else:
        #plt.show()
        return fig, ax

def plot_rv(IVobject, temps="all", chans="all", lgcsave=False, savepath="", savename=""):
    """
    Function to plot the resistance curves for the data in an IV object.

    Parameters
    ----------
    IVobject : class
        The IV class object that the data is stored in.
    temps : string, array_like, int, optional
        Which bath temperatures to plot. Setting to "all" plots all of them. Can also set
        to a subset of bath temperatures, or just one
    chans : string, array_like, int, optional
        Which bath temperatures to plot. Setting to "all" plots all of them. Can also set
        to a subset of bath temperatures, or just one
    lgcsave : boolean, optional
        Boolean flag to save the plot
    savepath : string, optional
        Path to save the plot to, saves it to the current directory by default
    savename : string, optional
        Name to append to the plot file name, if saving

    """

    ntemps, nch, niters = IVobject.r0.shape
    chan_names = IVobject.chan_names

    if temps == "all":
        trange = range(ntemps)
    elif np.isscalar(temps):
        trange = np.array([temps])
    else:
        trange = np.array(temps)

    if chans == "all":
        chrange = range(nch)
    elif np.isscalar(chans):
        chrange = np.array([chans])
    else:
        chrange = np.array(chans)

    ch_colors = plt.cm.viridis(np.linspace(0, 1, num=len(trange) * len(chrange)))

    fig, ax = plt.subplots(figsize=(10, 6))

    for it, t in enumerate(trange):
        for ich, ch in enumerate(chrange):
            if ntemps > 1:
                label_str="Temp {}, Channel {}".format(t, chan_names[ch])
            else:
                label_str="Channel {}".format(chan_names[ch])
            ax.scatter(
                IVobject.vb[t, ch] * 1e6,
                IVobject.r0[t, ch] * 1e3,
                label=label_str,
                color=ch_colors[it * len(chrange) + ich],
                s=10.0,
            )
            ax.plot(
                IVobject.vb[t, ch] * 1e6,
                IVobject.r0[t, ch] * 1e3,
                color=ch_colors[it * len(chrange) + ich],
                alpha=0.5,
            )
            ax.errorbar(
                IVobject.vb[t, ch] * 1e6,
                IVobject.r0[t, ch] * 1e3,
                yerr=IVobject.r0_err[t, ch] * 1e3,
                linestyle='None',
                color='k',
            )

    ax.legend(loc='best')
    ax.set_xlabel(r'Bias Voltage [$\mu V$]')
    ax.set_ylabel(r'Resistance of TES [$m \Omega$]')
    ax.grid(linestyle='dotted')
    ax.tick_params(which='both', direction='in', right=True, top=True)
    ax.set_title(r"$R_0$ vs. $V_b$")

    if lgcsave:
        fig.savefig(savepath + "rv_curve_{}.png".format(savename))
        plt.close(fig)
    else:
        #plt.show()
        return fig, ax    

def plot_pv(IVobject, temps="all", chans="all", lgcsave=False, savepath="", savename=""):
    """
    Function to plot the power curves for the data in an IV object.

    Parameters
    ----------
    IVobject : class
        The IV class object that the data is stored in.
    temps : string, array_like, int, optional
        Which bath temperatures to plot. Setting to "all" plots all of them. Can also set
        to a subset of bath temperatures, or just one
    chans : string, array_like, int, optional
        Which bath temperatures to plot. Setting to "all" plots all of them. Can also set
        to a subset of bath temperatures, or just one
    lgcsave : boolean, optional
        Boolean flag to save the plot
    savepath : string, optional
        Path to save the plot to, saves it to the current directory by default
    savename : string, optional
        Name to append to the plot file name, if saving

    """

    ntemps, nch, niters = IVobject.ptes.shape
    chan_names = IVobject.chan_names

    if temps == "all":
        trange = range(ntemps)
    elif np.isscalar(temps):
        trange = np.array([temps])
    else:
        trange = np.array(temps)

    if chans == "all":
        chrange = range(nch)
    elif np.isscalar(chans):
        chrange = np.array([chans])
    else:
        chrange = np.array(chans)

    ch_colors = plt.cm.viridis(np.linspace(0, 1, num=len(trange) * len(chrange)))

    fig, ax = plt.subplots(figsize=(10, 6))

    for it, t in enumerate(trange):
        for ich, ch in enumerate(chrange):
            if ntemps > 1:
                label_str="Temp {}, Channel {}".format(t, chan_names[ch])
            else:
                label_str="Channel {}".format(chan_names[ch])
            ax.scatter(
                IVobject.vb[t, ch] * 1e6,
                IVobject.ptes[t, ch] * 1e12,
                label=label_str,
                color=ch_colors[it * len(chrange) + ich],
                s=10.0,
            )
            ax.plot(
                IVobject.vb[t, ch] * 1e6,
                IVobject.ptes[t, ch] * 1e12,
                color=ch_colors[it * len(chrange) + ich],
                alpha=0.5,
            )
            ax.errorbar(
                IVobject.vb[t, ch] * 1e6,
                IVobject.ptes[t, ch] * 1e12,
                yerr=IVobject.ptes_err[t, ch] * 1e12,
                linestyle='None',
                color='k',
            )

    ax.legend(loc='best')
    ax.set_xlabel(r'Bias Voltage [$\mu V$]')
    ax.set_ylabel(r'Power [$pW$]')
    ax.grid(linestyle='dotted')
    ax.tick_params(which='both', direction='in', right=True, top=True)
    ax.set_title("$P_0$ vs. $V_b$")

    if lgcsave:
        fig.savefig(savepath + "pv_curve_{}.png".format(savename))
        plt.close(fig)
    else:
        #plt.show()
        return fig, ax

def plot_all_curves(IVobject, temps="all", chans="all", showfit=True, lgcsave=False, savepath="", savename=""):
    """
    Function to plot the IV, resistance, and power curves for the data in an IV object.

    Parameters
    ----------
    IVobject : class
        The IV class object that the data is stored in.
    temps : string, array_like, int, optional
        Which bath temperatures to plot. Setting to "all" plots all of them. Can also set
        to a subset of bath temperatures, or just one
    chans : string, array_like, int, optional
        Which bath temperatures to plot. Setting to "all" plots all of them. Can also set
        to a subset of bath temperatures, or just one
    showfit : boolean, optional
        Boolean flag to also plot the linear fit to the normal data
    lgcsave : boolean, optional
        Boolean flag to save the plot
    savepath : string, optional
        Path to save the plot to, saves it to the current directory by default
    savename : string, optional
        Name to append to the plot file name, if saving

    """

    plot_iv(
        IVobject,
        temps=temps,
        chans=chans,
        showfit=showfit,
        lgcsave=lgcsave,
        savepath=savepath,
        savename=savename,
    )
    plot_rv(
        IVobject,
        temps=temps,
        chans=chans,
        lgcsave=lgcsave,
        savepath=savepath,
        savename=savename,
    )
    plot_pv(
        IVobject,
        temps=temps,
        chans=chans,
        lgcsave=lgcsave,
        savepath=savepath,
        savename=savename,
    )




import numpy as np
from numpy import pi
from scipy.optimize import least_squares, fsolve
from scipy.fftpack import fft, ifft, fftfreq

from iminuit import Minuit
from ._base_didv import stdcomplex, complexadmittance, _BaseDIDV
from ._didv import didvinitfromdata
from ._plot_didv import _PlotDIDV


__all__ = [
    "DIDVPriors",
]


class DIDVPriors(_BaseDIDV, _PlotDIDV):
    """
    Class for fitting a didv curve for different types of models of the
    didv given prior information known about the parameters. This class
    supports doing 1, 2, and 3 pole fits. This is supported in a way
    that does one dataset at a time.

    """

    def __init__(self, rawtraces, fs, sgfreq, sgamp, rsh, tracegain=1.0,
                 dutycycle=0.5, add180phase=False, dt0=10.0e-6,
                 autoresample=False):
        """
        Initialization of the DIDVPriors class object

        Parameters
        ----------
        rawtraces : ndarray
            The array of rawtraces to use when fitting the didv. Should
            be of shape (number of traces, length of trace in bins).
            This can be any units, as long as tracegain will convert
            this to Amps.
        fs : float
            Sample rate of the data taken, in Hz
        sgfreq : float
            Frequency of the signal generator, in Hz
        sgamp : float
            Peak-to-peak size of the square wave supplied by the signal
            generator, in Amps (equivalent to jitter in the QET bias)
        rsh : float
            Expected shunt resistance in the circuit, Ohms
        tracegain : float, optional
            The factor that the rawtraces should be divided by to
            convert the units to Amps. If rawtraces already has units
            of Amps, then this should be set to 1.0
        dutycycle : float, optional
            The duty cycle of the signal generator, should be a float
            between 0 and 1. Set to 0.5 by default
        add180phase : boolean, optional
            If the signal generator is out of phase (i.e. if it looks
            like --__ instead of __--), then this should be set to
            True. Adds half a period of the signal generator to the dt0
            attribute
        dt0 : float, optional
            The value of the starting guess for the time offset of the
            didv when fitting. See Notes for more information.
        autoresample : bool, optional
            If True, the initialization will automatically resample
            the data so that `fs` / `sgfreq` is an integer, which
            ensures that an arbitrary number of signal-generator
            periods can fit in an integer number of time bins. See
            `qetpy.utils.resample_data` for more info.

        Notes
        -----
        The best way to use the `dt0` value if it isn't converging well
        is to run the fit multiple times, setting dt0 equal to the
        fit's next value, and seeing where the `dt0` value converges.
        The fit can have a difficult time finding the value on the
        first run if it the initial value is far from the actual value,
        so a solution is to do this iteratively.

        """

        super().__init__(
            rawtraces,
            fs,
            sgfreq,
            sgamp,
            rsh,
            tracegain=tracegain,
            dutycycle=dutycycle,
            add180phase=add180phase,
            dt0=dt0,
            autoresample=autoresample,
        )


    @staticmethod
    def _onepolescaledadmittance(freq, rsh, rp, L):
        """
        Function to calculate the admittance (didv), scaled by rsh,
        of a TES with the 1-pole (normal and SC) fit from Irwin's TES
        parameters. This is the functional form of dI_sensor/dI_bias.

        """

        didv = complexadmittance(freq, rsh=rsh, rp=rp, L=L)

        return rsh * didv


    @staticmethod
    def _twopolescaledadmittance(freq, rsh, rp, r0, beta, l, L, tau0):
        """
        Function to calculate the admittance (didv), scaled by rsh,
        of a TES with the 2-pole fit from Irwin's TES parameters. This
        is the functional form of the dI_sensor/dI_bias.

        """

        didv = complexadmittance(
            freq, rsh=rsh, rp=rp, r0=r0, beta=beta, l=l, L=L, tau0=tau0,
        )

        return rsh * didv


    @staticmethod
    def _threepolescaledadmittance(freq, rsh, rp, r0, beta, l, L, tau0,
                                   gratio, tau3):
        """
        Function to calculate the admittance (didv), scaled by rsh,
        of a TES with the 3-pole fit from Irwin's TES parameters. This
        is the functional form of the dI_sensor/dI_bias.

        """

        didv = complexadmittance(
            freq,
            rsh=rsh,
            rp=rp,
            r0=r0,
            beta=beta,
            l=l,
            L=L,
            tau0=tau0,
            gratio=gratio,
            tau3=tau3,
        )

        return rsh * didv

        
    @staticmethod
    def _fitdidv(freq, didv, poles, priors, invpriorscov, p0, yerr=None):
        """
        Function to directly fit the small signal TES parameters with
        the knowledge of prior known values any number of the
        parameters. In order for the degeneracy of the parameters to be
        broken, at least 2 fit parameters should have priors knowledge.
        This is usually rsh, rp, and r0, as these can be known from IV
        data.

        """

        def _residual(params):
            """
            Define a residual for the nonlinear least squares algorithm
            for the priors fit.

            """

            if poles == 1:
                rsh, rp, L, dt = params
                ci = DIDVPriors._onepolescaledadmittance(
                    freq, rsh, rp, L,
                ) * np.exp(-2.0j*pi*freq*dt)
            elif poles == 2:
                rsh, rp, r0, beta, l, L, tau0, dt = params
                ci = DIDVPriors._twopolescaledadmittance(
                    freq, rsh, rp, r0, beta, l, L, tau0,
                ) * np.exp(-2.0j*pi*freq*dt)
            elif poles == 3:
                rsh, rp, r0, beta, l, L, tau0, gratio, tau3, dt = params
                ci = DIDVPriors._threepolescaledadmittance(
                    freq, rsh, rp, r0, beta, l, L, tau0, gratio, tau3,
                ) * np.exp(-2.0j*pi*freq*dt)

            # the difference between the data and the fit
            diff = didv - ci
            # get the weights from yerr, these should be
            # 1/(standard deviation) for real and imaginary parts
            if (yerr is None):
                weights = 1.0+1.0j
            else:
                weights = 1.0 / yerr.real + 1.0j / yerr.imag

            # create the residual vector, splitting up real and
            # imaginary parts of the residual separately
            z1d = np.zeros(freq.size*2, dtype = np.float64)
            z1d[0:z1d.size:2] = diff.real*weights.real
            z1d[1:z1d.size:2] = diff.imag*weights.imag
            return z1d


        def _residualpriors(params):
            """Helper function to incude the priors in the residual."""

            z1dpriors = np.sqrt(
                (
                    priors - params
                ).dot(
                    invpriorscov
                ).dot(
                    priors - params
                )
            )
            return z1dpriors

        def _neg_log_likelihood(params):
            """Negative log likelihood with priors included."""

            return np.sum(
                (_residual(params))**2 / 2
            ) + _residualpriors(params)**2 / 2

        m = Minuit(
            _neg_log_likelihood,
            p0,
        )
        m.limits = (len(p0) - 1) * ((0, None), ) + ((None, None), )
        m.errors = np.abs(p0)
        m.errordef = 0.5

        m.migrad()

        popt = np.asarray(m.values)
        pcov = np.asarray(m.covariance)
        cost = m.fval

        return popt, pcov, cost


    def _guessparams(self, poles, fcutoff, priors):
        """
        Hidden method for using the non-priors fitting to guess the
        starting parameters.

        """

        DIDVGuess = didvinitfromdata(
            self._tmean,
            self._didvmean,
            self._didvstd,
            self._offset,
            self._offset_err,
            self._fs,
            self._sgfreq,
            self._sgamp,
            self._rsh,
            add180phase=self._add180phase,
            dt0=self._dt0,
            rp=priors[1] if priors[1] != 0 else 0.01,
            r0=priors[2] if priors[2] != 0 else 0.1,
        )

        DIDVGuess.dofit(poles, fcutoff=fcutoff)

        dt = DIDVGuess.fitresult(poles)['params']['dt']
        ssparams = DIDVGuess.fitresult(poles)['smallsignalparams']

        if poles == 1:
            p0 = (
                ssparams['rsh'],
                ssparams['rp'],
                ssparams['L'],
                dt,
            )
        elif poles == 2:
            p0 = (
                ssparams['rsh'],
                ssparams['rp'],
                ssparams['r0'],
                ssparams['beta'],
                ssparams['l'],
                ssparams['L'],
                ssparams['tau0'],
                dt,
            )
        elif poles == 3:
            p0 = (
                ssparams['rsh'],
                ssparams['rp'],
                ssparams['r0'],
                ssparams['beta'],
                ssparams['l'],
                ssparams['L'],
                ssparams['tau0'],
                ssparams['gratio'],
                ssparams['tau3'],
                dt,
            )
        else:
            raise ValueError("The number of poles should be 1, 2, or 3.")

        return (*np.abs(p0[:-1]), p0[-1])


    def dofit(self, poles, priors, priorscov, fcutoff=np.inf):
        """
        Class function to fit dIdV using the inputted priors. For best
        results, provide as many priors and corresponding prior
        covariance entries as possible.

        Parameters
        ----------
        poles : int
            The fit that should be run. Should be 1, 2, or 3.
        priors : ndarray
            A 1-d array that contains the priors parameters for the
            specified fit. See Notes for more information.
        priorscov : ndarray
            An 2-d array that contains the priors covariance matrix for
            the specified fit. See Notes for more information.
        fcutoff : float, optional
            The cutoff frequency in Hz, above which data is ignored in
            the specified fitting routine. Default is `np.inf`, which
            is equivalent to no cutoff frequency.

        Notes
        -----
        The `priors` and `priorscov` have specific forms that should be
        followed for each model.

            For the 1-pole model, the priors array should be a 4-entry
            array of the form (rsh, rp, L, dt), and the covariance
            matrix should be a 4-by-4 array that describes the known
            entries in the covariance matrix.

            For the 2-pole model, the priors array should be an 8-entry
            array of the form (rsh, rp, r0, beta, l, L, tau0, dt), and
            the covariance matrix should be an 8-by-8 array that
            describes the known entries in the covariance matrix.

            For the 2-pole model, the priors array should be a 10-entry
            array of the form (rsh, rp, r0, beta, l, L, tau0, gratio,
            tau3, dt), and the covariance matrix should be a 10-by-10
            array that describes the known entries in the covariance
            matrix.

        """

        if self._tmean is None:
            self.processtraces()

        fit_freqs = np.abs(self._freq) < fcutoff

        guess = self._guessparams(poles, fcutoff, priors)

        guess_new = [g if p == 0 else p for g, p in zip(guess, priors)]

        params, cov, cost = DIDVPriors._fitdidv(
            self._freq[fit_freqs],
            self._didvmean[fit_freqs] * self._rsh,
            poles,
            priors,
            np.linalg.pinv(priorscov),
            p0=guess_new,
            yerr=self._didvstd[fit_freqs] * self._rsh,
        )

        falltimes = DIDVPriors._findpolefalltimes(
            DIDVPriors._convertfromtesvalues(params),
        )

        if poles == 1:
            self._1poleresult = DIDVPriors._fitresult(
                poles, params, cov, falltimes, cost,
            )
            self._1poleresult['priors'] = priors
            self._1poleresult['priorscov'] = priorscov
        elif poles == 2:
            self._2poleresult = DIDVPriors._fitresult(
                poles, params, cov, falltimes, cost,
            )
            self._2poleresult['priors'] = priors
            self._2poleresult['priorscov'] = priorscov
        elif poles == 3:
            self._3poleresult = DIDVPriors._fitresult(
                poles, params, cov, falltimes, cost,
            )
            self._3poleresult['priors'] = priors
            self._3poleresult['priorscov'] = priorscov


    @staticmethod
    def _fitresult(poles, params, cov, falltimes, cost):
        """
        Function for converting data from different fit results to a
        results dictionary.

        """

        result = dict()

        if poles == 1:
            result['params'] = {
                'rsh': params[0],
                'rp': params[1],
                'L': params[2],
                'dt': params[3],
            }
            result['cov'] = cov
            errors = np.diag(cov)**0.5
            result['errors'] = {
                'rsh': errors[0],
                'rp': errors[1],
                'L': errors[2],
                'dt': errors[3],
            }

        if poles == 2:
            result['params'] = {
                'rsh': params[0],
                'rp': params[1],
                'r0': params[2],
                'beta': params[3],
                'l': params[4],
                'L': params[5],
                'tau0': params[6],
                'dt': params[7],
            }
            result['cov'] = cov
            errors = np.diag(cov)**0.5
            result['errors'] = {
                'rsh': errors[0],
                'rp': errors[1],
                'r0': errors[2],
                'beta': errors[3],
                'l': errors[4],
                'L': errors[5],
                'tau0': errors[6],
                'dt': errors[7],
            }

        if poles == 3:
            result['params'] = {
                'rsh': params[0],
                'rp': params[1],
                'r0': params[2],
                'beta': params[3],
                'l': params[4],
                'L': params[5],
                'tau0': params[6],
                'gratio': params[7],
                'tau3': params[8],
                'dt': params[9],
            }
            result['cov'] = cov
            errors = np.diag(cov)**0.5
            result['errors'] = {
                'rsh': errors[0],
                'rp': errors[1],
                'r0': errors[2],
                'beta': errors[3],
                'l': errors[4],
                'L': errors[5],
                'tau0': errors[6],
                'gratio': errors[7],
                'tau3': errors[8],
                'dt': errors[9],
            }

        result['falltimes'] = falltimes
        result['cost'] = cost
        result['didv0'] = complexadmittance(0, **result['params']).real

        return result



from . import didv
from .didv import *
from ._iv import *
from ._ibis import *
from ._noise import *
from ._fitting import *
from ._of_nsmb import *
from ._of_pileup import *
from ._de_pileup import *
from ._of_base import *
from ._of_1x1 import *
from ._of_nonlin import *



import numpy as np
from scipy import interpolate, signal, constants
from scipy import ndimage
from sympy.ntheory import factorrat
from sympy.core.symbol import S


__all__ = [
    "make_decreasing",
    "calc_offset",
    "lowpassfilter",
    "align_traces",
    "get_offset_from_muon",
    "powertrace_simple",
    "energy_absorbed",
    "slope",
    "fill_negatives",
    "shift",
    "make_template",
    "estimate_g",
    "resample_factors",
    "resample_data",
    "interpolate_parabola",
    "interpolate_of",
    "argmin_chisq",
]


def shift(arr, num, fill_value=0):
    """
    Function for shifting the values in an array by a certain number of
    indices, filling the values of the bins at the head or tail of the
    array with fill_value.

    Parameters
    ----------
    arr : array_like
        Array to shift values in.
    num : float
        The number of values to shift by. If positive, values shift to
        the right. If negative, values shift to the left. If num is a
        non-whole number of bins, arr is linearly interpolated
    fill_value : scalar, optional
        The value to fill the bins at the head or tail of the array with.

    Returns
    -------
    result : ndarray
        The resulting array that has been shifted and filled in.

    """

    result = np.empty_like(arr)

    if float(num).is_integer():
        num = int(num) # force num to int type for slicing

        if num > 0:
            result[:num] = fill_value
            result[num:] = arr[:-num]
        elif num < 0:
            result[num:] = fill_value
            result[:num] = arr[-num:]
        else:
            result[:] = arr
    else:
        result = ndimage.shift(
            arr, num, order=1, mode='constant', cval=fill_value,
        )

    return result


def resample_factors(fs, sgfreq):
    """
    Function for determining the upsampling and downsampling factors
    needed to ensure that `fs`/`sgfreq` is an integer. These factor are
    to be used with `scipy.signal.resample_poly`. Note that these
    factors are not a unique solution, but a simple one.

    Parameters
    ----------
    fs : int
        The digitization rate of the data in Hz.
    sgfreq : int
        The frequency of the square wave from the signal generator in
        Hz.

    Returns
    -------
    up : int
        The upsampling factor.
    down : int
        The downsampling factor.

    Notes
    -----
    See the documentation for `scipy.signal.resample_poly` for more
    information on these factors.

    """

    if int(fs) != fs:
        raise ValueError('`fs` must be an integer to use this function')
    if int(sgfreq) != sgfreq:
        raise ValueError('`sgfreq` must be an integer to use this function')

    rfacs = factorrat(S(int(fs)) / int(sgfreq))
    numer = [key for key in rfacs if rfacs[key] > 0]
    denom = [key for key in rfacs if rfacs[key] < 0]

    if len(denom)==0:
        return 1, 1

    down = np.multiply.reduce(numer)
    up = round(down / min(denom)) * min(denom)

    # round up if `sgfreq` is prime (otherwise would return 0)
    if up==0:
        up = np.ceil(down / min(denom)).astype(int) * min(denom)

    return int(up), int(down)


def resample_data(x, fs, sgfreq, **kwargs):
    """
    Function that uses `resample_factors` and
    `scipy.signal.resample_poly` to automatically resample the data to
    ensure that `fs`/`sgfreq` is an integer.

    Parameters
    ----------
    x : array_like
        The data to be resampled.
    fs : int
        The digitization rate of the data in Hz.
    sgfreq : int
        The frequency of the square wave from the signal generator in
        Hz.
    axis : int, optional
        The axis of `x` that is resampled. Default is -1.
    window : string, tuple, or array_like, optional
        Desired window to use to design the low-pass filter, or the FIR
        filter coefficients to employ.
    padtype : string, optional
        `constant`, `line`, `mean`, `median`, `maximum`, `minimum` or
        any of the other signal extension modes supported by
        `scipy.signal.upfirdn`. Changes assumptions on values beyond
        the boundary. If `constant`, assumed to be `cval` (default
        zero). If `line` assumed to continue a linear trend defined by
        the first and last points. `mean`, `median`, `maximum` and
        `minimum` work as in `np.pad` and assume that the values beyond
        the boundary are the mean, median, maximum or minimum
        respectively of the array along the axis. Default is
        `constant`.
    cval : float, optional
        Value to use if `padtype='constant'`. Default is zero.

    Returns
    -------
    resampled_x : ndarray
        The resampled data.
    resampled_fs : float
        The digitization rate of `resampled_x` in Hz.

    Notes
    -----
    The `kwargs` are each passed to the `scipy.signal.resample_poly`
    function, from which we took the text for this dosctring for those
    parameters.

    """

    if 'axis' not in kwargs:
        kwargs['axis'] = -1

    up, down = resample_factors(fs, sgfreq)
    resampled_x = signal.resample_poly(x, up, down, **kwargs)
    resampled_fs =  fs * up / down

    return resampled_x, resampled_fs


def make_template(t, tau_r, tau_f, offset=0):
    """
    Function to make an ideal pulse template in time domain with single
    pole exponential rise and fall times, and a given time offset. The
    template will be returned with the maximum pulse height normalized
    to one. The pulse, by default, begins at the center of the trace,
    which can be left or right shifted via the `offset` optional
    argument.

    Parameters
    ----------
    t : ndarray
        Array of time values to make the pulse with
    tau_r : float
        The time constant for the exponential rise of the pulse
    tau_f : float
        The time constant for the exponential fall of the pulse
    offset : int
        The number of bins the pulse template should be shifted

    Returns
    -------
    template_normed : array
        the pulse template in time domain

    """

    pulse = np.exp(-t/tau_f)-np.exp(-t/tau_r)
    pulse_shifted = shift(pulse, len(t)//2 + offset)
    template_normed = pulse_shifted/pulse_shifted.max()

    return template_normed

def make_decreasing(y, x=None):
    """
    Function to take an array of values and make it monotonically
    decreasing. This is done by simply tossing out any values that are
    larger than the last value, moving from the first index to the
    last, and interpolating between values.

    Parameters
    ----------
    y : ndarray
        Array of values to make monotonically decreasing.
    x : ndarray, optional
        The x-values corresponding to `y`, can be useful if the
        x-values are not evenly spaced.

    Returns
    -------
    out : ndarray
        A monotonically decreasing version of `y`.

    """

    if x is None:
        x = np.arange(len(y))

    y_dec = np.zeros(len(y))
    y_dec[0] = y[0]
    last_val = y[0]

    for ii, val in enumerate(y):
        if (last_val > val):
            last_val = y[ii]
            y_dec[ii] = y[ii]

    interp_inds = y_dec!=0

    f = interpolate.interp1d(
        x[interp_inds], y[interp_inds], fill_value="extrapolate",
    )

    out = f(x)

    return out


def calc_offset(x, fs=1.0, sgfreq=100.0, is_didv=False):
    """
    Calculates the DC offset of time series trace.

    Parameters
    ----------
    x : ndarray
        Array to calculate offsets of.
    fs : float, optional
        Sample rate of the data being taken, assumed to be in units
        of Hz.
    sgfreq : float, optional
        The frequency of signal generator (if is_didv is True. If False,
        then this is ignored).
    is_didv : bool, optional
        If False, average of full trace is returned. If True, then the
        average of n periods is returned (where n is the max number of
        full periods present in a trace).

    Returns
    -------
    offset : ndarray
        Array of offsets with same shape as input x minus the last
        dimension.
    std : ndarray
        Array of std with same shape as offset.

    """

    if is_didv:
        period =  1.0 / sgfreq
        period_bins = period * fs
        n_periods = int(x.shape[-1] / period_bins)
        x = x[..., :int(n_periods * period_bins)]

    offset = np.mean(np.mean(x, axis=-1), axis=0)
    std = np.std(np.mean(x, axis=-1), axis=0) / np.sqrt(x.shape[0])

    return offset, std

def slope(x, y, removemeans=True):
    """
    Computes the maximum likelihood slope of a set of x and y points.

    Parameters
    ----------
    x : array_like
        Array of real-valued independent variables.
    y : array_like
        Array of real-valued dependent variables.
    removemeans : bool
        Boolean flag for if the mean of x should be subtracted. This
        should be set to True if x has not already had its mean
        subtracted. Set to False if the mean has been subtracted.
        Default is True.

    Returns
    -------
    slope : float
        Maximum likelihood slope estimate, calculated as
        sum((x-<x>)(y-<y>))/sum((x-<x>)**2)

    """

    x_mean = np.mean(x) if removemeans else 0

    return np.sum((x - x_mean) * (y - x_mean)) / np.sum((x - x_mean) ** 2)

def fill_negatives(arr):
    """
    Simple helper function to remove negative and zero values from PSDs
    and replace them with interpolated values.

    Parameters
    ----------
    arr : ndarray
        Array of values to replace neagive values on

    Returns
    -------
    arr : ndarray
        Modified input array with the negative and zero values replace
        by interpolated values.

    """

    zeros = np.array(arr <= 0)
    inds_zero = np.where(zeros)[0]
    inds_not_zero = np.where(~zeros)[0]
    good_vals = arr[~zeros]

    if len(good_vals) != 0:
        arr[zeros] = np.interp(inds_zero, inds_not_zero, good_vals)

    return arr



def lowpassfilter(traces, cut_off_freq=100000, fs=625e3, order=1):
    """
    Applies a low pass filter to the inputted time series traces.

    Parameters
    ----------
    traces : ndarray
        An array of shape (# traces, # bins per trace).
    cut_off_freq : float, int, optional
        The cut off 3dB frequency for the low pass filter, defaults to
        100000 Hz.
    fs : float, int, optional
        Digitization rate of data, defaults to 625e3 Hz.
    order : int, optional
        The order of the low pass filter, defaults to 1.

    Returns
    -------
    filt_traces : ndarray
        Array of low pass filtered traces with the same shape as
        inputted traces.

    """

    nyq = 0.5 * fs
    cut_off = cut_off_freq / nyq
    b,a = signal.butter(order, cut_off)
    filt_traces = signal.filtfilt(b, a, traces, padtype='even')

    return filt_traces

def align_traces(traces, lgcjustshifts=False, n_cut=5000, cut_off_freq=5000.0,
                 fs=625e3):
    """
    Function to align dIdV traces if each trace does not trigger at the
    same point. Uses a convolution of the traces to find the time
    offset.

    Parameters
    ----------
    traces : ndarray
        Array of shape (# traces, # bins per trace).
    lgcjustshifts : boolean, optional
        If False, the aligned traces and the phase shifts are returned.
        If True, just the phase shifts are returned. Default is False.
    n_cut : int, optional
        The number of bins to use to do the convolution. Just need
        enough information to see the periodic signal. Default is
        5000.
    cut_off_freq : float or int, optional
        3dB cut off frequency for filter. Default is 5000 Hz.
    fs : float or int, optional
        Sample rate of data in Hz. Default is 625e3 Hz.

    Returns
    -------
    shifts : ndarray
        Array of phase shifts for each trace in units of bins.
    masked_aligned : masked ndarray, optional
        Array of time shift corrected traces, same shape as input
        traces. The masked array masks the np.NaN values in the time
        shifted traces so that normal numpy functions will ignore the
        nan's in computations.

    """

    # Filter and truncate all traces to speed up.
    traces_filt = lowpassfilter(
        traces[:, :n_cut], cut_off_freq=5000, fs=625e3,
    )
    traces_temp = traces_filt - np.mean(traces_filt, axis=-1, keepdims=True)
    traces_norm = traces_temp / (np.amax(traces_temp, axis=-1, keepdims=True))

    # use the first trace to define the origin of alignment
    t1 = traces_norm[0]
    # define the origin
    orig = np.argmax(signal.fftconvolve(t1,t1[::-1],mode = 'full'))

    # initialize empty array to store the aligned traces
    traces_aligned = np.zeros_like(traces)
    shifts = np.zeros(traces.shape[0])

    for ii in range(traces.shape[0]):
        t2 = traces_norm[ii]
        # Convolve each trace against the origin trace, find the index of the
        # max value, then subtract of the index of the origin trace
        t2_shift = np.argmax(
            signal.fftconvolve(t1, t2[::-1], mode='full'),
        ) - orig
        shifts[ii] = t2_shift
        if not lgcjustshifts:
            traces_aligned[ii] = ndimage.shift(
                traces[ii], t2_shift, cval=np.nan,
            )

    if lgcjustshifts:
        return shifts
    else:
        flat_aligned = traces_aligned.flatten()
        masked_aligned = np.ma.array(
            flat_aligned,
            mask=np.isnan(flat_aligned),
        ).reshape(traces_aligned.shape)

        return shifts, masked_aligned

def get_offset_from_muon(avemuon, qetbias, rn, rload, rsh=5e-3,
                         nbaseline=6000, lgcfullrtn=True):
    """
    Function to calculate the offset in the measured TES current using
    an average muon.

    Parameters
    ----------
    avemuon : ndarray
        An average of 'good' muons in time domain, referenced to TES
        current.
    qetbias : float
        Applied QET bias current
    rn : float
        Normal resistance of the TES
    rload : float
        Load resistance of TES circuit (rp + rsh)
    rsh : float, optional
        Value of the shunt resistor for the TES circuit
    nbaseline : int, optional
        The number of bins to use to calculate the baseline,
        i.e. [0:nbaseline].
    lgcfullrtn : bool, optional
        If True, the offset, r0, i0, and bias power is returned. If
        False, just the offset is returned.

    Returns
    -------
    ioffset : float
        The offset in the measured TES current.
    r0 : float, optional
        The resistance of the TES.
    i0 : float, optional
        The quiescent current through the TES.
    p0 : float, optional
        The quiescent bias power of the TES.

    """

    muon_max = np.max(avemuon)
    baseline = np.mean(avemuon[:int(nbaseline)])
    peak_loc = np.argmax(avemuon)

    muon_saturation = np.mean(avemuon[peak_loc:peak_loc+200])
    muon_deltaI =  muon_saturation - baseline

    vbias = qetbias * rsh
    inormal = vbias / (rload + rn)

    i0 = inormal - muon_deltaI
    ioffset = baseline - i0

    r0 = inormal * (rn + rload)/i0 - rload
    p0 = i0 * rsh * qetbias - rload * i0**2

    if lgcfullrtn:
        return ioffset, r0, i0, p0
    else:
        return ioffset

def powertrace_simple(trace, ioffset, qetbias, rload, rsh):
    """
    Function to convert time series trace from units of TES current to
    units of power. This can be done for either a single trace, or an
    array of traces, as long as the first dimension is the number of
    traces.

    The function takes into account the second order dependence on
    current, but assumes the infinite Irwin loop gain approximation.

    Parameters
    ----------
    trace : ndarray
        Time series traces, where the last dimension is the trace
        length, referenced to TES current.
    ioffset : float
        The offset in the measured TES current.
    qetbias : float
        Applied QET bias current.
    rload : float
        Load resistance of TES circuit (rp + rsh).
    rsh : float
        Value of the shunt resistor for the TES circuit.

    Returns
    -------
    trace_p : ndarray
        Time series trace, in units of power referenced to the TES.

    """

    vbias = qetbias * rsh
    trace_i0 = trace - ioffset
    trace_p = trace_i0*vbias - (rload) * trace_i0**2

    return trace_p


def energy_absorbed(trace, ioffset, qetbias, rload, rsh, fs=None,
                    baseline=None, time=None, indbasepre=None,
                    indbasepost=None):
    """
    Function to calculate the energy collected by the TESs by
    integrating the power in the TES as a function of time. This can be
    done for either a single trace, or an array of traces, as long as
    the first dimension is the number of traces.

    Parameters
    ----------
    trace : ndarray
        Time series traces, where the last dimension is the trace
        length, referenced to TES current.
    ioffset : float
        The offset in the measured TES current.
    qetbias : float
        Applied QET bias current.
    rload : float
        Load resistance of TES circuit (rp + rsh).
    rsh : float
        Value of the shunt resistor for the TES circuit.
    fs : float, optional
        The sample rate of the DAQ
    baseline : ndarray, optional
        The baseline value of each trace, must be same dimension as
        trace.
    time : ndarray, optional
        Array of time values corresponding to the trace array.
    indbasepre : int, optional
        The bin number corresponding to the pre-pulse baseline,
        i.e. [:indbasepre]
    indbasepost : int, optional
        The bin number corresponding to the post-pulse baseline,
        i.e. [indbasepost:]

    Returns
    -------
    integrated_energy : float, ndarray
        The energy absorbed by the TES in units of eV.

    """

    if baseline is None:
        if indbasepre is not None:
            base_traces = trace[..., :indbasepre]
        else:
            raise ValueError('Must provide indbasepre or baseline')
        if indbasepost is not None:
            base_traces = np.hstack((base_traces, trace[..., indbasepost:]))
        baseline = np.mean(base_traces, axis=-1, keepdims=True)

    baseline_p0 = powertrace_simple(baseline, ioffset, qetbias, rload, rsh)
    trace_power = powertrace_simple(trace, ioffset, qetbias, rload, rsh)

    if fs is not None:
        integrated_energy = np.trapz(
            baseline_p0 - trace_power, axis=-1,
        ) / (fs * constants.e)
    elif time is not None:
        integrated_energy = np.trapz(
            baseline_p0 - trace_power, x=time, axis=-1,
        ) / constants.e
    else:
        raise ValueError('Must provide either fs or time')

    return integrated_energy

def estimate_g(p0, tc, tbath, p0_err=0, cov=None, n=5):
    """
    Function to estimate G given the measured bias power and know Tc
    and bath temperature.

    Parameters
    ----------
    p0 : float
        The applied bias power.
    tc : float
        The SC transition temperature of the TES.
    tbath : float
        The bath temperature.
    p0_err : float, optional
        The error in the bias power.
    cov : ndarray, NoneType, optional
        The covariance matrix for the parameters in order: p0, tc,
        tbath. If None, the error is just calculated from p0_err.
    n : int, optional
        The exponent of the power law expression. Defaults to 5.

    Returns
    -------
    g : float
        The estimated thermal conductance.
    g_err : float
        The error in the estimated thermal conductance.

    """

    g = n * p0 * tc**(n - 1) / (tc**n - tbath**n)

    if cov is not None:
        dgdp = g / p0
        dgdtc = (n - 1) * g / tc - g * n * (tc**(n - 1)) / (tc**n - tbath**n)
        dgdtbath = n * g * tbath**(n - 1) / (tc**n - tbath**n)

        jac = np.zeros((3,3))
        jac[0,0] = dgdp
        jac[1,1] = dgdtc
        jac[2,2] = dgdtbath
        covout = jac.dot(cov.dot(jac.transpose()))
        g_err = np.sqrt(np.sum(covout))
    else:
        g_err = np.abs(p0_err * g / p0)

    return g, g_err




def interpolate_parabola(vals, bestind, delta, t_interp=None):
    """
    Precomputed equation of a parabola given 3 equally spaced
    points. Returns the coordinates of the extremum of the
    parabola.
    """

    sf = 1 / (2 * delta**2)
    
    a = sf * (vals[bestind + 1] - 2 * vals[bestind] + vals[bestind - 1])
    b = sf * delta * (vals[bestind + 1] - vals[bestind - 1])
    c = sf * 2 * delta**2 * vals[bestind]
    
    if t_interp is None:
        t_interp = - b / (2 * a)
    vals_interp = a * t_interp**2 + b * t_interp + c
        
    return t_interp, vals_interp


    
   
def interpolate_of(amps, chi2, bestind, delta):
    """
    Helper function for running `_interpolate_parabola` twice,
    in the correct order.
    """
    
    t_interp, chi2_interp = interpolate_parabola(
        chi2, bestind, delta,
    )
    _, amps_interp = interpolate_parabola(
        amps, bestind, delta, t_interp=t_interp,
    )
    
    return amps_interp, t_interp, chi2_interp




def argmin_chisq(chisq,
                 window_min=None,
                 window_max=None,
                 lgc_outside_window=False,
                 constraint_mask=None):
    """
    Helper function for finding the index for the minimum of a chi^2.
    Includes options for constraining the values of chi^2.
        
    Parameters
    ----------
    chi2 : ndarray
             An array containing the chi^2 to minimize. If `chi2` has
             dimension greater than 1, then it is minimized along the last
             axis. 

    window_min : NoneType, int, optional
         This is the window minimum length  (in bins) to
          constrain the possible values to in the chi^2 minimization,
          Default is None, `chi2` is uncontrained is both 
          window min/max = None
    
    window_max : NoneType, int, optional
          This is the window maximum length  (in bins) to
          constrain the possible values to in the chi^2 minimization,
          Default is None, `chi2` is uncontrained is both 
          window min/max = None
    

    
    lgcoutsidewindow : bool, optional
        If False, then the function will minimize the chi^2 in the bins
        inside the constrained window specified by window_min/max.
        If True, the function will minimize the chi^2 in the bins outside 
        window
        
    constraint_mask : NoneType, boolean ndarray, optional
        An additional constraint on the chi^2 to apply, which should be
        in the form of a boolean mask. If left as None, no additional
        constraint is applied.
    
    Returns
    -------
    bestind : int, ndarray, float
        The index of the minimum of `chi2` given the constraints
        specified by `nconstrain` and `lgcoutsidewindow`. If the
        dimension of `chi2` is greater than 1, then this will be an
        ndarray of ints.
    
    """

    # intitialize
    bestind = np.nan
    
    # number samples
    nbins = len(chisq)

    # case constraints
    if (window_min is not None
        or window_max is not None):
        
        # check window min
        if  window_min is None or window_min<0:
            window_min = 0
                    
        # check window max
        if  window_max is None or window_max>nbins:
            window_max = nbins
          

        if window_min>window_max:
            raise ValueError('ERROR: OF window min bin bigger than window max bin!')


        inds=[]
        if lgc_outside_window:
            inds = np.concatenate(
                (np.arange(0, window_min),
                 np.arange(window_max, nbins))
            )
        else:
            inds = np.arange(window_min, window_max)

        if len(inds)==0:
            raise ValueError('ERROR: OF window is empty. Check arguments')
        
        inds = inds[(inds>=0) & (inds<nbins)]
            
        if constraint_mask is not None:
            inds = inds[constraint_mask[inds]]
            
        if len(inds)!=0:
            bestind = np.argmin(chisq[..., inds], axis=-1)
            bestind = inds[bestind]

    else:

        if constraint_mask is None:
            bestind = np.argmin(chisq, axis=-1)
        else:
            inds = np.flatnonzero(constraint_mask)
            if len(inds)!=0:
                bestind = np.argmin(chisq[..., constraint_mask],
                                    axis=-1)
                bestind = inds[bestind]
               

    return bestind



import os
import numpy as np
import matplotlib.pyplot as plt
from qetpy import autocuts, IV, Noise, calc_psd
from qetpy.sim import TESnoise
from qetpy.plotting import compare_noise, plot_noise_sim

THIS_DIR = os.path.dirname(os.path.abspath(__file__))

def test_autocuts():
    pathtodata = os.path.join(THIS_DIR, "data/test_autocuts_data.npy")
    traces = np.load(pathtodata)
    fs = 625e3
    cut = autocuts(traces, fs=fs)
    assert len(cut)>0

def test_iv():
    pathtodata = os.path.join(THIS_DIR, "data/test_iv_data.npz")
    testdata = np.load(pathtodata)
    dites = testdata["dites"]
    dites_err = testdata["dites_err"]
    vb = testdata["vb"]
    vb_err = testdata["vb_err"]
    rload = testdata["rload"]
    rload_err = testdata["rload_err"]

    ivdata = IV(dites, dites_err, vb, vb_err, rload, rload_err, ["A","B","C"])
    ivdata.calc_iv()

    ivdata.plot_all_curves()

    assert len(ivdata.r0) > 0

def test_noise():
    pathtodata = os.path.join(THIS_DIR, "data/traces.npy")
    traces_PT_on = np.load(pathtodata)

    savePath = '' #user needs to define new path

    sampleRate = 625e3 #define sample rate
    channels = ['PCS1', 'PES1', 'PFS1', 'PAS2', 'PBS2', 'PES2', 'PDS2'] #define the channel names
    g124_noise = Noise(traces_PT_on, sampleRate, channels) #initialize a noise object
    g124_noise.name = 'G124 SLAC Run 37 Pulse Tube On'

    g124_noise.calculate_psd()
    g124_noise.calculate_corrcoeff()
    g124_noise.calculate_csd()

    g124_noise.calculate_uncorr_noise()
    g124_noise.plot_psd(lgcoverlay=True)
    g124_noise.plot_corrcoeff(lgcsave=False, lgcsmooth=True, nwindow=13)
    g124_noise.plot_csd(whichcsd=['66','26'])
    g124_noise.plot_reim_psd()
    g124_noise.plot_decorrelatednoise(lgccorrelated=True)

    noise_sim = TESnoise(freqs=g124_noise.freqs[1:])

    plot_noise_sim(
        f=g124_noise.freqs,
        psd=g124_noise.psd[0, :],
        noise_sim=noise_sim,
        istype='power',
        qetbias=1,
        lgcsave=False,
        figsavepath='',
        xlims=None,
        ylims=None,
    )

    plot_noise_sim(
        f=g124_noise.freqs,
        psd=g124_noise.psd[0, :],
        noise_sim=noise_sim,
        istype='current',
        qetbias=1,
        lgcsave=False,
        figsavepath='',
        xlims=None,
        ylims=None,
    )

    plot_noise_sim(
        f=g124_noise.freqs,
        psd=g124_noise.psd[0, :],
        noise_sim=noise_sim,
        istype='sc',
        qetbias=1,
        lgcsave=False,
        figsavepath='',
        xlims=None,
        ylims=None,
    )

    plot_noise_sim(
        f=g124_noise.freqs,
        psd=g124_noise.psd[0, :],
        noise_sim=noise_sim,
        istype='normal',
        qetbias=1,
        lgcsave=False,
        figsavepath='',
        xlims=None,
        ylims=None,
    )

    plot_noise_sim(
        f=g124_noise.freqs,
        psd=g124_noise.psd[0, :],
        noise_sim=noise_sim,
        istype='power',
        qetbias=1,
        lgcsave=False,
        figsavepath='',
        xlims=(10, 1e5),
        ylims=(1e-12, 1e-9),
    )

    plot_noise_sim(
        f=g124_noise.freqs,
        psd=g124_noise.psd[0, :],
        noise_sim=noise_sim,
        istype='current',
        qetbias=1,
        lgcsave=False,
        figsavepath='',
        xlims=(10, 1e5),
        ylims=(1e-12, 1e-9),
    )

    plot_noise_sim(
        f=g124_noise.freqs,
        psd=g124_noise.psd[0, :],
        noise_sim=noise_sim,
        istype='sc',
        qetbias=1,
        lgcsave=False,
        figsavepath='',
        xlims=(10, 1e5),
        ylims=(1e-12, 1e-9),
    )

    plot_noise_sim(
        f=g124_noise.freqs,
        psd=g124_noise.psd[0,:],
        noise_sim=noise_sim,
        istype='normal',
        qetbias=1,
        lgcsave=False,
        figsavepath='',
        xlims=(10, 1e5),
        ylims=(1e-12, 1e-9),
    )


    assert len(g124_noise.psd) > 0
    assert len(noise_sim.freqs) > 0




import qetpy as qp
import numpy as np
import pytest


def _initialize_didv(poles, priors, sgfreq=100, autoresample=False):
    """Function for initializing dIdV data"""
    np.random.seed(0)

    rsh = 5e-3
    rbias_sg = 20000
    fs = 625e3
    sgamp = 0.009381 / rbias_sg

    rfb = 5000
    loopgain = 2.4
    drivergain = 4
    adcpervolt = 65536 / 2
    tracegain = rfb * loopgain * drivergain * adcpervolt

    true_params = {
        'rsh': rsh,
        'rp': 0.006,
        'r0': 0.0756 if poles in [2, 3] else 0,
        'beta': 2 if poles in [2, 3] else 0,
        'l': 10 if poles in [2, 3] else 0,
        'L': 1e-7,
        'tau0': 500e-6 if poles in [2, 3] else 0,
        'gratio': 0.5 if poles in [3] else 0,
        'tau3': 1e-3 if poles in [3] else 0,
    }

    psd_test = np.ones(int(4 * fs / sgfreq)) / tracegain**2 / 1e4
    rawnoise = qp.gen_noise(psd_test, fs=fs, ntraces=300)

    t = np.arange(rawnoise.shape[-1]) / fs
    didv_response = qp.squarewaveresponse(
        t, sgamp, sgfreq, **true_params,
    )
    rawtraces = didv_response + rawnoise

    if priors:
        if poles == 1:
            dim = 4
        elif poles == 2:
            dim = 8
        elif poles == 3:
            dim = 10
        priors = np.zeros(dim)
        priorscov = np.zeros((dim, dim))

        priors[0] = true_params['rsh']
        priorscov[0, 0] = (0.1 * priors[0])**2
        priors[1] = true_params['rp']
        priorscov[1, 1] = (0.1 * priors[1])**2

        if poles != 1:
            priors[2] = true_params['r0']
            priorscov[2, 2] = (0.1 * priors[2])**2

        didvfit = qp.DIDVPriors(
            rawtraces,
            fs,
            sgfreq,
            sgamp,
            rsh,
            tracegain=1.0,
            dt0=-1e-6,
            autoresample=autoresample,
        )
        if poles == 1:
            didvfit.processtraces()
            didvfit.plot_full_trace()
        assert didvfit.fitresult(poles) == dict()
        didvfit.dofit(poles, priors, priorscov)
    else:
        didvfit = qp.DIDV(
            rawtraces,
            fs,
            sgfreq,
            sgamp,
            rsh,
            tracegain=1.0,
            r0=true_params['r0'],
            rp=true_params['rp'],
            dt0=-1e-6 - 1 / (2 * sgfreq),
            add180phase=True,
            autoresample=autoresample,
        )
        assert didvfit.fitresult(poles) == dict()
        didvfit.dofit(poles)

    _run_plotting_suite(didvfit, poles)

    return didvfit, true_params


def _run_plotting_suite(didvfit, poles):
    """Helper function for running all plots."""

    didvfit.plot_full_trace(poles=poles)
    didvfit.plot_full_trace(saveplot=True, savename='test')
    didvfit.plot_single_period_of_trace()
    didvfit.plot_single_period_of_trace(saveplot=True, savename='test')
    didvfit.plot_didv_flipped()
    didvfit.plot_didv_flipped(saveplot=True, savename='test')
    didvfit.plot_zoomed_in_trace()
    didvfit.plot_zoomed_in_trace(saveplot=True, savename='test')
    didvfit.plot_abs_phase_didv()
    didvfit.plot_abs_phase_didv(saveplot=True, savename='test')
    didvfit.plot_re_vs_im_dvdi(poles=poles)
    didvfit.plot_re_vs_im_dvdi(saveplot=True, savename='test')
    didvfit.plot_re_im_didv()
    didvfit.plot_re_im_didv(saveplot=True, savename='test')


def test_errors():
    """
    Function for asserting certain errors are raised
    for specific cases.

    """

    error_str = "`fs` and `sgfreq` do not divide to an integer."

    with pytest.raises(ValueError) as excinfo:
        didvfit = qp.DIDV(
            None,
            625e3, # fs
            30, # sgfreq
            None,
            None,
        )

    assert error_str in str(excinfo.value)


def test_autoresample():
    """
    Function for testing the autoresample kwarg for _BaseDIDV.

    """

    np.random.seed(0)

    sgfreq = 90
    rsh = 5e-3
    rbias_sg = 20000
    fs = 625e3
    sgamp = 0.009381 / rbias_sg

    rawtraces = np.random.rand(300, 32768)

    didvfit = qp.DIDV(
        rawtraces,
        fs,
        sgfreq,
        sgamp,
        rsh,
        autoresample=True,
    )

    expected_length = np.ceil(rawtraces.shape[-1] * 9 / 10)

    assert didvfit._rawtraces.shape[-1] == expected_length


def test_didv():
    """Function for testing the DIDV and DIDVPriors classes."""

    poles = [1, 2, 3]
    priors = [True, False]
    keys = ['params', 'smallsignalparams']

    for pole in poles:
        for key, prior in zip(keys, priors):
            didvfit, true_params = _initialize_didv(pole, prior)
            assert np.isclose(
                qp.complexadmittance(
                    1e4, **didvfit.fitresult(pole)[key],
                ),
                qp.complexadmittance(
                    1e4, **true_params,
                ),
                rtol=1e-2,
            )




import numpy as np
from scipy.optimize import least_squares
from numpy.fft import rfft, fft, ifft, fftfreq, rfftfreq
from qetpy.plotting import plotnonlin
from qetpy.utils import shift


__all__ = ["OFnonlin"]



class OFnonlin(object):
    """
    This class provides the user with a non-linear optimum filter to
    estimate the amplitude, rise time (optional), fall time, and time
    offset of a pulse.

    Attributes:
    -----------
    psd : ndarray
        The power spectral density corresponding to the pulses that
        will be used in the fit. Must be the full psd (positive and
        negative frequencies), and should be properly normalized to
        whatever units the pulses will be in.
    fs : int or float
        The sample rate of the ADC
    df : float
        The delta frequency
    freqs : ndarray
        Array of frequencies corresponding to the psd
    time : ndarray
        Array of time bins corresponding to the pulse
    template : ndarray
        The time series pulse template to use as a guess for initial
        parameters
    data : ndarray
        FFT of the pulse that will be used in the fit
    lgcdouble : bool
        If False, only the Pulse hight, fall time, and time offset will
        be fit. If True, the rise time of the pulse will be fit in
        addition to the above.
    taurise : float
        The user defined risetime of the pulse
    error : ndarray
        The uncertianty per frequency (the square root of the psd,
        divided by the errorscale)
    dof : int
        The number of degrees of freedom in the fit
    norm : float
        Normalization factor to go from continuous to FFT
    scale_amplitude : bool
        If using the 1- or 2-pole fit, whether the parameter, A, should
        be treated as the pulse height (`scale_amplitude` = True,
        default) or as a scale parameter in the functional expression.
        See `twopole` and `twopoletime` for details.

    """

    def __init__(self, psd, fs, template=None):
        """
        Initialization of OFnonlin object

        Parameters
        ----------
        psd : ndarray
            The power spectral density corresponding to the pulses that
            will be used in the fit. Must be the full psd (positive and
            negative frequencies), and should be properly normalized to
            whatever units the pulses will be in.
        fs : int, float
            The sample rate of the ADC
        template : ndarray, NoneType, optional
            The time series pulse template to use as a guess for
            initial parameters, if inputted.

        """

        self.psd = np.zeros(len(psd))
        self.psd[:] = psd
        self.psd[0] = 1e40

        self.fs = fs
        self.df = fs / len(psd)
        self.freqs = np.fft.fftfreq(len(psd), 1 / fs)
        self.time = np.arange(len(psd)) / fs
        self.template = template

        self.data = None
        self.npolefit = 1
        self.scale_amplitude = True

        self.taurise = None
        self.error = None
        self.dof = None
        self.norm = np.sqrt(fs * len(psd))


    def fourpole(self, A, B, C, tau_r, tau_f1, tau_f2, tau_f3, t0):
        """
        Functional form of pulse in frequency domain with 1 rise time
        and three fall times. The fall times have independent
        amplitudes (A,B,C). The condition f(0)=0 requires the rise time
        to have amplitude (A+B+C). Therefore, the "amplitudes" take on
        different meanings than in other n-pole functions. The
        functional form (time-domain) is:

            A*(exp(-t/tau_fall1)) + B*(exp(-t/tau_fall2)) +
            C*(exp(-t/tau_fall3)) - (A+B+C)*(exp(-t/tau_rise))

        4 rise/fall times, 3 amplitudes, and time offset allowed to
        float.

        Parameters
        ----------
        A : float
            Amplitude for first fall time
        B : float
            Amplitude for second fall time
        C : float
            Amplitude for third fall time
        tau_r : float
            Rise time of pulse
        tau_f1 : float
            First fall time of pulse
        tau_f2 : float
            Second fall time of pulse
        tau_f3 : float
            Third fall time of pulse
        t0 : float
            Time offset of four pole pulse

        Returns
        -------
        pulse : ndarray, complex
            Array of amplitude values as a function of frequency

        """

        omega = 2 * np.pi * self.freqs
        phaseTDelay = np.exp(-(0 + 1j) * omega * t0)
        pulse = (
            (
                A * (tau_f1 / (1 + omega * tau_f1 * (0 + 1j)))
            ) + (
                B * (tau_f2 / (1 + omega * tau_f2 * (0 + 1j)))
            ) + (
                C * (tau_f3 / (1 + omega * tau_f3 * (0 + 1j)))
            ) - (
                (A + B + C) * (tau_r / (1 + omega * tau_r * (0 + 1j)))
            )
        ) * phaseTDelay
        return pulse * np.sqrt(self.df)

    def fourpoletime(self, A, B, C, tau_r, tau_f1, tau_f2, tau_f3, t0):
        """
        Functional form of pulse in time domain with 1 rise time and
        three fall times The fall times have independent amplitudes
        (A,B,C). The condition f(0)=0 requires the rise time to have
        amplitude (A+B+C). Therefore, the "amplitudes" take on
        different meanings than in other n-pole functions. The
        functional form (time-domain) is:

            A*(exp(-t/tau_fall1)) + B*(exp(-t/tau_fall2)) +
            C*(exp(-t/tau_fall3)) - (A+B+C)*(exp(-t/tau_rise))

        4 rise/fall times, 3 amplitudes, and time offset allowed to
        float.

        Parameters
        ----------
        A : float
            Amplitude for first fall time
        B : float
            Amplitude for second fall time
        C : float
            Amplitude for third fall time
        tau_r : float
            Rise time of pulse
        tau_f1 : float
            First fall time of pulse
        tau_f2 : float
            Second fall time of pulse
        tau_f3 : float
            Third fall time of pulse
        t0 : float
            Time offset of four pole pulse

        Returns
        -------
        pulse : ndarray
            Array of amplitude values as a function of time

        """

        pulse = (
            A * (np.exp(-self.time / tau_f1))
        ) + (
            B * (np.exp(-self.time / tau_f2))
        ) + (
            C * (np.exp(-self.time / tau_f3))
        ) - (
            (A + B + C) * (np.exp(-self.time / tau_r))
        )
        return shift(pulse, int(t0 * self.fs))

    def threepole(self, A, B, tau_r, tau_f1, tau_f2, t0):
        """
        Functional form of pulse in frequency domain with 1 rise time
        and two fall times. The  fall times have independent amplitudes
        (A,B) and the condition f(0)=0 constrains the rise time to have
        amplitude (A+B). The functional form (time domain) is:

            A*(exp(-t/\tau_fall1)) + B*(exp(-t/\tau_fall2)) -
            (A+B)*(exp(-t/\tau_rise))

        and therefore the "amplitudes" take on different meanings than
        in the other n-pole functions

        3 rise/fall times, 2 amplitudes, and time offset allowed to
        float.

        Parameters
        ----------
        A : float
            Amplitude for first fall time
        B : float
            Amplitude for second fall time
        tau_r : float
            Rise time of pulse
        tau_f1 : float
            First fall time of pulse
        tau_f2 : float
            Second fall time of pulse
        t0 : float
            Time offset of three pole pulse

        Returns
        -------
        pulse : ndarray, complex
            Array of amplitude values as a function of frequency

        """

        omega = 2 * np.pi * self.freqs
        phaseTDelay = np.exp(-(0 + 1j) * omega * t0)
        pulse = (
            (
                A * (tau_f1 / (1 + omega * tau_f1 * (0 + 1j)))
            ) + (
                B * (tau_f2 / (1 + omega * tau_f2 * (0 + 1j)))
            ) - (
                (A + B) * (tau_r / (1 + omega * tau_r * (0 + 1j)))
            )
        ) * phaseTDelay
        return pulse * np.sqrt(self.df)


    def threepoletime(self, A, B, tau_r, tau_f1, tau_f2, t0):
        """
        Functional form of pulse in time domain with 1 rise time and
        two fall times. The  fall times have independent amplitudes
        (A,B) and the condition f(0)=0 constrains the rise time to have
        amplitude (A+B). The functional form (time domain) is:

            A*(exp(-t/\tau_fall1)) + B*(exp(-t/\tau_fall2)) - 
            (A+B)*(exp(-t/\tau_rise))

        and therefore the "amplitudes" take on different meanings than
        in the other n-pole functions

        3 rise/fall times, 2 amplitudes, and time offset allowed to
        float.

        Parameters
        ----------
        A : float
            Amplitude for first fall time
        B : float
            Amplitude for second fall time
        tau_r : float
            Rise time of pulse
        tau_f1 : float
            First fall time of pulse
        tau_f2 : float
            Second fall time of pulse
        t0 : float
            Time offset of three pole pulse

        Returns
        -------
        pulse : ndarray
            Array of amplitude values as a function of time

        """

        pulse = (
            A * (np.exp(-self.time / tau_f1))
        ) + (
            B * (np.exp(-self.time / tau_f2))
        ) - (
            (A + B) * (np.exp(-self.time / tau_r))
        )
        return shift(pulse, int(t0 * self.fs))


    def twopole(self, A, tau_r, tau_f, t0):
        """
        Functional form of pulse in frequency domain with the
        amplitude, rise time, fall time, and time offset allowed to
        float. The functional form (time domain) is:

            A*(exp(-t/\tau_fall)) - A*(exp(-t/\tau_rise))

        Note that there are 2 ways to interpret the 'A' parameter input
        to this function (see below).

        This is meant to be a private function

        Parameters
        ----------
        A : float
            Amplitude paramter or pulse height. If self.scale_amplitude
            is true, A represents the pulse height, if false, A is the
            amplitude parameter in the time domain expression above.
        tau_r : float
            Rise time of two-pole pulse
        tau_f : float
            Fall time of two-pole pulse
        t0 : float
            Time offset of two-pole pulse

        Returns
        -------
        pulse : ndarray, complex
            Array of amplitude values as a function of frequency

        """

        omega = 2 * np.pi * self.freqs

        if(self.scale_amplitude):
            delta = tau_r - tau_f
            rat = tau_r / tau_f
            amp = A / (rat**(-tau_r / delta) - rat**(-tau_f / delta))
            pulse = amp * np.abs(
                tau_r-tau_f
            ) / (
                1 + omega * tau_f * 1j
            ) / (
                1 + omega * tau_r * 1j
            ) * np.exp(-omega * t0 * 1.0j)
        else:
            pulse = (
                (
                    A * (tau_f / (1 + omega * tau_f * (0 + 1j)))
                ) - (
                    A * (tau_r / (1 + omega * tau_r * (0 + 1j)))
                )
            ) * np.exp(-omega * t0 * 1.0j)

        return pulse * np.sqrt(self.df)



    def twopoletime(self, A, tau_r, tau_f, t0):
        """
        Functional form of pulse in time domain with the amplitude,
        rise time, fall time, and time offset allowed to float. The
        functional form (time domain) is:

            A*(exp(-t/\tau_fall)) - A*(exp(-t/\tau_rise))

        Note that there are 2 ways to interpret the 'A' parameter input
        to this function (see below).

        This is meant to be a private function

        Parameters
        ----------
        A : float
            Amplitude paramter or pulse height. If self.scale_amplitude
            is true, A represents the pulse height, if false, A is the
            amplitude parameter in the time domain expression above.
        tau_r : float
            Rise time of two-pole pulse
        tau_f : float
            Fall time of two-pole pulse
        t0 : float
            Time offset of two pole pulse

        Returns
        -------
        pulse : ndarray
            Array of amplitude values as a function of time

        """

        if(self.scale_amplitude):
            delta = tau_r - tau_f
            rat = tau_r / tau_f
            amp = A / (rat**(-tau_r / delta) - rat**(-tau_f / delta))
            pulse = amp * (
                np.exp(-(self.time) / tau_f) - np.exp(-(self.time) / tau_r)
            )
        else:
            pulse = (
                A * (np.exp(-self.time / tau_f))
            ) - (
                A * (np.exp(-self.time / tau_r))
            )

        return shift(pulse, int(t0 * self.fs))


    def onepole(self, A, tau_f, t0):
        """
        Functional form of pulse in time domain with the amplitude,
        fall time, and time offset allowed to float, and the rise time
        held constant

        Parameters
        ----------
        A : float
            Amplitude of pulse
        tau_f : float
            Fall time of two-pole pulse
        t0 : float
            Time offset of two pole pulse

        Returns
        -------
        pulse : ndarray, complex
            Array of amplitude values as a function of frequency

        """

        tau_r = self.taurise
        return self.twopole(A, tau_r, tau_f, t0)


    
    def residuals(self, params):
        """
        Function to calculate the weighted residuals to be minimized

        Parameters
        ----------
        params : tuple
            Tuple containing fit parameters

        Returns
        -------
        z1d : ndarray
            Array containing residuals per frequency bin. The complex
            data is flatted into a single array

        """

        if (self.npolefit==4):
            A, B, C, tau_r, tau_f1, tau_f2, tau_f3, t0 = params
            delta = (self.data - self.fourpole(
                A, B, C, tau_r, tau_f1, tau_f2, tau_f3, t0,
            ))
        elif (self.npolefit==3):
            A, B, tau_r, tau_f1, tau_f2, t0 = params
            delta = (self.data - self.threepole(
                A, B, tau_r, tau_f1, tau_f2, t0,
            ))
        elif (self.npolefit==2):
            A, tau_r, tau_f, t0 = params
            delta = (self.data - self.twopole(
                A, tau_r, tau_f, t0,
            ))
        else:
            A, tau_f, t0 = params
            delta = (self.data - self.onepole(A, tau_f, t0))
            
        z1d = np.zeros(self.data.size * 2, dtype=np.float64)
        z1d[0:z1d.size:2] = delta.real / self.error
        z1d[1:z1d.size:2] = delta.imag / self.error

        return z1d




    def calcchi2(self, model):
        """
        Function to calculate the reduced chi square

        Parameters
        ----------
        model : ndarray
            Array corresponding to pulse function (twopole or onepole)
            evaluated at the optimum values

        Returns
        -------
        chi2 : float
            The reduced chi squared statistic

        """

        return sum(
            np.abs(self.data - model)**2 / self.error**2
        ) / (
            len(self.data) - self.dof
        )




    def fit_falltimes(self, pulse, npolefit=1, errscale=1, guess=None,
                      bounds=None, lgcfix=None, taurise=None, scale_amplitude=True,
                      lgcfullrtn=True, lgcplot=False, verbose=0):

        """"
        Renamed to dofit, keep for back compatibility
        """
      
        return self.dofit(pulse,
                          npolefit=npolefit,
                          errscale=errscale,
                          guess=guess,
                          bounds=bounds,
                          lgcfix=lgcfix,
                          taurise=taurise,
                          scale_amplitude=scale_amplitude,
                          lgcfullrtn=lgcfullrtn,
                          lgcplot=lgcplot,
                          verbose=verbose)
    

    
    def dofit(self, pulse, npolefit=1, errscale=1, guess=None,
              bounds=None, lgcfix=None, taurise=None, scale_amplitude=True,
              lgcfullrtn=True, lgcplot=False, verbose=0):
        """
        Function to do the fit

        Parameters
        ----------
        pulse : ndarray
            Time series traces to be fit. Should be a 1-dimensional
            array.
        npolefit: int, optional
            The number of poles to fit.
            If 1, the one pole fit is done, the user must provide the
            value of taurise
            If 2, the two pole fit is done
            If 3, the three pole fit is done (1 rise 2 fall). Second
            fall time amplitude is independent
            If 4, the four pole fit is done (1 rise 3 fall). Second and
            third fall time amplitudes are independent
        errscale : float or int, optional
            A scale factor for the psd. For example, if fitting an
            average, the errscale should be set to the number of traces
            used in the average.
        guess : 1d numpy array, optional
            Guess of initial values for fit, must be the same size as
            the model being used for fit.
        bounds : 2-tuple of 1D numpy array , optional
            Lower and upper bounds on independent variables. Each array
            must match the size of guess. Use np.inf with an
            appropriate sign to disable bounds on all or some
            variables. If None, bounds are automatically set to within
            a factor of 100 of amplitude guesses, a factor of 10 of
            rise/fall time guesses, and within 30 samples of start time
            guess.
        lgcfix : 1D numpy array (boolean)
            array size of nb parameters. If True, fix parameter using guess value
        taurise : float, optional
            The value of the rise time of the pulse if the single pole
            function is being use for fit
        scale_amplitude : bool, optional
            If using the 1- or 2-pole fit, whether the parameter, A,
            should be treated as the pulse height 
            (`scale_amplitude`=True, default) or as a scale parameter
            in the functional expression. See `twopole` and
            `twopoletime` for details.
        lgcfullrtn : bool, optional
            If False, only the best fit parameters are returned. If
            True, the errors in the fit parameters, the covariance
            matrix, and chi squared statistic are returned as well.
        lgcplot : bool, optional
            If True, diagnostic plots are returned.

        Returns
        -------
        variables : tuple
            The best fit parameters
        errors : tuple, optional
            The corresponding fit errors for the best fit parameters.
            Returned if `lgcfullrtn` is True.
        cov : ndarray, optional
            The convariance matrix returned from the fit. Returned if
            `lgcfullrtn` is True.
        chi2 : float, optional
            The reduced chi squared statistic evaluated at the optimum
            point of the fit. Returned if `lgcfullrtn` is True.
        success : bool, optional
           The success flag from `scipy.optimize.curve_fit`. True if
           the fit converged. Returned if `lgcfullrtn` is True.

        Raises
        ------
        ValueError
            if length of guess does not match the number of parameters
            needed in fit

        """

        # FFT pulse 
        self.data = np.fft.fft(pulse) / self.norm
        self.error = np.sqrt(self.psd / errscale)

        self.npolefit = npolefit
        self.scale_amplitude = scale_amplitude

        if (self.npolefit==1):
            if taurise is None:
                raise ValueError(
                    'taurise must not be None if doing 1-pole fit.'
                )
            else:
                self.taurise = taurise



        # initial guess
        p0 = None
        
        # before making guesses, if self.template
        # has been defined then define maxind,
        # ampscale, and amplitudes using the template.
        # otherwise use the pulse
        if self.template is not None:
            ampscale = np.max(pulse) - np.min(pulse)
            templateforguess = self.template
        else:
            ampscale = 1
            templateforguess = pulse

        maxind = np.argmax(templateforguess)

        # 4-pole fit
        if self.npolefit==4:

            self.dof = 8

            
            # guesses need to be tuned depending
            # on the detector being analyzed.
            # good guess for t0 particularly important to provide
            Aguess = np.mean(
                templateforguess[maxind - 7:maxind + 7]
            ) * ampscale
            Bguess = Aguess / 3
            Cguess = Aguess / 3
            tauriseguess = 20e-6
            taufall1guess = 100e-6
            taufall2guess = 300e-6
            taufall3guess = 500e-6
            t0guess = maxind / self.fs


            # replace by user guess
            if guess is not None:

                # check length
                if len(guess) != 8:
                    raise ValueError(
                        "Length of guess not compatible with 4-pole fit. "
                        "Must be of format: guess = (A,B,C,taurise,taufall1,"
                        "taufall2,taufall3,t0)"
                    )

                # guessed values
                (Aguess_in, Bguess_in, Cguess_in, tauriseguess_in, taufall1guess_in,
                 taufall2guess_in, taufall3guess_in, t0guess_in,) = guess

                if Aguess_in is not None:
                    Aguess = Aguess_in
                if Bguess_in is not None:
                    Bguess = Bguess_in  
                if Cguess_in is not None:
                    Cguess = Cguess_in  
                if tauriseguess_in:
                    tauriseguess =  tauriseguess
                if taufall1guess_in  is not None:
                    taufall1guess = taufall1guess_in
                if taufall2guess_in  is not None:
                    taufall2guess = taufall2guess_in
                if taufall3guess_in  is not None:
                    taufall3guess = taufall3guess_in  
                if t0guess_in is not None:
                    t0guess = t0guess_in


            # p0 / boounds
            p0 = np.array((Aguess, Bguess, Cguess, tauriseguess,
                           taufall1guess, taufall2guess,
                           taufall3guess, t0guess))
            
            if bounds is None:
                boundslower = np.array(
                    [Aguess / 100, Bguess / 100, Cguess / 100,
                     tauriseguess / 10, taufall1guess / 10,
                     taufall2guess / 10, taufall3guess / 10,
                     t0guess - 300 / self.fs]
                )
                boundsupper =  np.array(
                    [Aguess * 100, Bguess * 100, Cguess * 100,
                     tauriseguess * 10, taufall1guess * 10,
                     taufall2guess * 10, taufall3guess * 10,
                     t0guess + 300 / self.fs]
                )

                bounds = (boundslower, boundsupper)
                

        # 3-pole fit
        elif self.npolefit==3:
            
            self.dof = 6

            # default guess
            Aguess = np.mean(
                templateforguess[maxind - 7:maxind + 7]
            ) * ampscale
            Bguess = Aguess / 3
            tauriseguess = 20e-6
            taufall1guess = 100e-6
            taufall2guess = 300e-6
            t0guess = maxind / self.fs


            # replace by user guess
            if guess is not None:

                # check length
                if len(guess) != 6:
                    raise ValueError(
                        "Length of guess not compatible with 3-pole fit. "
                        "Must be of format: guess = (A,B,taurise,taufall1,"
                        "taufall2,t0)"
                    )

                # guessed values
                (Aguess_in, Bguess_in, tauriseguess_in, taufall1guess_in,
                 taufall2guess_in, t0guess_in) = guess
              
                if Aguess_in is not None:
                    Aguess = Aguess_in
                if Bguess_in is not None:
                    Bguess = Bguess_in  
                if tauriseguess_in:
                    tauriseguess =  tauriseguess
                if taufall1guess_in  is not None:
                    taufall1guess = taufall1guess_in
                if taufall2guess_in  is not None:
                    taufall2guess = taufall2guess_in
                if t0guess_in is not None:
                    t0guess = t0guess_in

            p0 = np.array((Aguess, Bguess, tauriseguess,
                           taufall1guess, taufall2guess,
                           t0guess))

            
            if bounds is None:
                boundslower = np.array(
                    [Aguess / 100, Bguess / 100, tauriseguess / 10,
                     taufall1guess / 10, taufall2guess / 10,
                     t0guess - 300 / self.fs]
                )
                boundsupper = np.array([
                    Aguess * 100, Bguess * 100, tauriseguess * 10,
                    taufall1guess * 10, taufall2guess * 10,
                    t0guess + 300 / self.fs]
                )
                
                bounds = (boundslower, boundsupper)
            
        elif self.npolefit==2:

            self.dof = 4

            # default guess
            ampguess = np.mean(
                templateforguess[maxind-7:maxind+7]
            ) * ampscale
            tauval = 0.37 * ampguess
            endt_val = int(300e-6 * self.fs)
            tauind = np.argmin(
                np.abs(
                    pulse[maxind + 1:maxind + 1 + endt_val] - tauval
                )
            ) + maxind + 1
            taufallguess = (tauind - maxind) / self.fs
            tauriseguess = 20e-6
            t0guess = maxind / self.fs


            # replace by user guess
            if guess is not None:

                # check length
                if len(guess) != 4:
                    raise ValueError(
                        'Length of guess not compatible with 2-pole fit. '
                        'Must be of format: guess = (A,taurise,taufall,t0)'
                    )
                
                ampguess_in, tauriseguess_in, taufallguess_in, t0guess_in = guess

                if ampguess_in is not None:
                    ampguess = ampguess_in
                if tauriseguess_in:
                    tauriseguess =  tauriseguess
                if taufallguess_in  is not None:
                    taufallguess = taufallguess_in
                if t0guess_in is not None:
                    t0guess = t0guess_in
                    
            p0 = np.array((ampguess, tauriseguess, taufallguess, t0guess))

            if bounds is None:
                boundslower = np.array(
                    [ampguess / 100, tauriseguess / 10,
                     taufallguess / 10, t0guess - 300 / self.fs]
                )         
                boundsupper =  np.array(
                    [ampguess * 100, tauriseguess * 10,
                     taufallguess * 10, t0guess + 300 / self.fs]
                )
                bounds = (boundslower, boundsupper)
        else:

            self.dof = 3
            
            # default guess
            ampguess = np.mean(
                templateforguess[maxind-7:maxind+7]
            ) * ampscale
            tauval = 0.37 * ampguess
            endt_val = int(300e-6 * self.fs)
            tauind = np.argmin(
                np.abs(
                    pulse[maxind + 1:maxind + 1 + endt_val] - tauval
                )
            ) + maxind + 1
            taufallguess = (tauind - maxind) / self.fs
            t0guess = maxind / self.fs


            # replace by user guess
            if guess is not None:

                # check
                if len(guess) != 3:
                    raise ValueError(
                        'Length of guess not compatible with 1-pole fit. '
                        'Must be of format: guess = (A,taufall,t0)'
                    )

                ampguess_in, taufallguess_in, t0guess_in = guess

                if ampguess_in is not None:
                    ampguess = ampguess_in
                if taufallguess_in  is not None:
                    taufallguess = taufallguess_in
                if t0guess_in is not None:
                    t0guess = t0guess_in
                    
            p0 = np.array((ampguess, taufallguess, t0guess))

            if bounds is None:
                boundslower = np.array(
                    [ampguess / 100, taufallguess / 10,
                     t0guess - 300 / self.fs]
                    )
                boundsupper = np.array(
                    [ampguess * 100, taufallguess * 10,
                     t0guess + 300 / self.fs]
                )
                bounds = (boundslower, boundsupper)



        # fix parameters
        fix_params = None
        if lgcfix is not None:

            fix_params =  p0[lgcfix].copy()
            
            p0 = p0[~lgcfix]
            bounds = (bounds[0][~lgcfix],
                      bounds[1][~lgcfix])
            
          
                

        def _residual_lsq(var_params):
            """
            Function that is passed to nonlinear 
            least_squares scipy algorithm
            
            Parameters
            ----------

            var_params : np.array (dtype=float)
               variable fit parameters array

            
            Return
            ------

            residual :  np.array
            """

            if fix_params is None:
                return self.residuals(var_params)
            else:
                all_params = np.zeros_like(lgcfix, dtype=float)
                np.place(all_params, lgcfix, fix_params)
                np.place(all_params, ~lgcfix, var_params)
                return self.residuals(all_params)


            
        # FIT
        result = least_squares(
            _residual_lsq,
            x0=p0,
            bounds=bounds,
            x_scale=p0,
            jac='3-point',
            loss='linear',
            xtol=2.3e-16,
            ftol=2.3e-16,
            verbose=verbose
        )
        
        variables = result['x'].copy()
        if lgcfix is not None:
            variables = np.zeros_like(lgcfix, dtype=float)
            np.place(variables, lgcfix, fix_params)
            np.place(variables, ~lgcfix, result['x'])
        
        success = result['success']
        cost = result['cost']
        jac = result['jac']
        cov = np.linalg.pinv(np.dot(np.transpose(jac), jac))
        errors = np.sqrt(cov.diagonal())


        # add back fixed variables
        variables = result['x'].copy()
        
        if lgcfix is not None:
            
            # parameters
            variables = np.zeros_like(lgcfix, dtype=float)
            np.place(variables, lgcfix, fix_params)
            np.place(variables, ~lgcfix, result['x'])

            #errors
            errors_all = np.zeros_like(lgcfix, dtype=float)
            np.place(errors_all, ~lgcfix, errors)
            errors = errors_all
            
        
        # chi2
        if self.npolefit==4:
            chi2 = self.calcchi2(
                self.fourpole(
                    variables[0],
                    variables[1],
                    variables[2],
                    variables[3],
                    variables[4],
                    variables[5],
                    variables[6],
                    variables[7],
                )
            )
        elif self.npolefit==3:
            chi2 = self.calcchi2(
                self.threepole(
                    variables[0],
                    variables[1],
                    variables[2],
                    variables[3],
                    variables[4],
                    variables[5],
                )
            )
        elif self.npolefit==2:
            chi2 = self.calcchi2(
                self.twopole(
                    variables[0],
                    variables[1],
                    variables[2],
                    variables[3],
                )
            )
        else:
            chi2 = self.calcchi2(
                self.onepole(
                    variables[0],
                    variables[1],
                    variables[2],
                )
            )

        if lgcplot:
            plotnonlin(self, pulse, variables, errors)

        if lgcfullrtn:
            return (variables, errors, cov, chi2, success)
        else:
            return variables


        





import qetpy as qp
import numpy as np
from scipy import stats
import pytest
from qetpy.cut._cut import _UnbiasedEstimators

def test_itercov():
    """Testing function for `qetpy.cut.itercov`."""

    np.random.seed(1)
    arr = np.random.rand(100, 3) + 4 * np.random.poisson(size=(100, 3))

    expected_res = (
        np.array(
            [4.20538738, 4.4593193, 3.63154184]
        ),
        np.array(
            [[11.83350126, 1.21053041, -0.04518929],
             [1.21053041, 9.70931534, 0.29405381],
             [-0.04518929,  0.29405381,  9.56345222]]
        ),
        np.array(
            [True, True, True, True, True, True, True, False, True,
             True, False, True, True, True, True, True, True, True,
             True, True, True, True, True, True, False, False, True,
             True, True, True, True, True, True, True, True, True,
             True, True, True, False, True, True, True, True, True,
             True, True, True, True, True, True, True, True, True,
             True, True, True, True, True, True, True, True, True,
             True, False, True, True, True, True, True, True, True,
             True, True, True, True, True, True, True, True, True,
             True, True, True, True, True, True, True, True, True,
             True, True, True, False, True, True, True, True, True,
             True]
        ),
    )

    res1 = qp.cut.itercov(arr)
    res2 = qp.cut.itercov(arr[:, 0], arr[:, 1], arr[:, 2])

    assert np.all(np.allclose(expected_res[ii], res1[ii]) for ii in range(3))
    assert np.all(np.all(res1[ii] == res2[ii]) for ii in range(3))

    with pytest.raises(ValueError):
        qp.cut.itercov(np.random.rand(10))

    with pytest.raises(ValueError):
        qp.cut.itercov(np.random.rand(10), np.random.rand(9))

    arr_in = np.array([[0, 1]])
    assert np.all(qp.cut.itercov(arr_in)[0] == arr_in[0])

def test_UnbiasedEstimators():
    """Testing function for `qetpy.cut._cut._UnbiasedEstimators`."""

    x = stats.norm.rvs(size=100, random_state=1)

    lwrbnd = -1
    uprbnd = 1

    unb = _UnbiasedEstimators(x, lwrbnd, uprbnd)

    assert np.allclose((unb.mu, unb.std), (-0.008724932112491217, 0.9540689615563553))

def test_iterstat():
    """Testing function for `qetpy.cut.iterstat`."""

    nsig = 3
    x = stats.norm.rvs(size=100, random_state=1)
    x[0] += 100
    res = qp.cut.iterstat(x, cut=nsig)

    expected_mask = np.ones(len(x), dtype=bool)
    expected_mask[0] = False
    expected_mu0 = x[expected_mask].mean()
    expected_std0 = x[expected_mask].std()
    expected_res = (expected_mu0, expected_std0, expected_mask)

    assert np.all([np.all(expected_res[ii] == res[ii]) for ii in range(3)])

    res_unb = qp.cut.iterstat(x, cut=nsig, return_unbiased_estimates=True)
    unb = _UnbiasedEstimators(x, expected_mu0 - nsig * expected_std0, expected_mu0 + nsig * expected_std0)
    expected_res_unb = (unb.mu, unb.std, expected_mask)

    assert np.all([np.all(expected_res_unb[ii] == res_unb[ii]) for ii in range(3)])

    expected_res_warn = (0, 0, np.array([True]))
    res_warn = qp.cut.iterstat(np.array([0]))

    assert np.all([np.all(expected_res_warn[ii] == res_warn[ii]) for ii in range(3)])



import numpy as np
import scipy.constants as constants


__all__ = [
    "energy_res_estimate",
    "loadfromdidv",
    "TESnoise",
]


def energy_res_estimate(freqs, tau_collect, Sp, collection_eff):
    """
    Function to estimate the energy resolution based on given noise and
    ideal pulse shape

    Parameters
    ----------
    freqs : array,
        Array of frequency values to integrate over
    tau_collect : float
        The collection time of the sensor
    Sp : array 
        Power spectral density (must be one-sided, see qetpy.foldpsd)
    collection_eff : float
        The collection efficiency of the detector

    Returns
    -------
    energy_res : float
        The estimated energy resolution in eV

    """

    omega = 2*np.pi*freqs
    single_pole = collection_eff/(1.+ omega*tau_collect*1j)
    integrand = 2*np.abs(single_pole)**2/(np.pi*Sp)
    energy_res = np.sqrt(1/np.trapz(integrand, x = omega))/constants.e

    return energy_res


def loadfromdidv(DIDVobj, G=5.0e-10, qetbias=160e-6, tc=0.040, tload=0.9,
                 tbath=0.020, squiddc=2.5e-12, squidpole=0.0, squidn=1.0,
                 rnormal=None, noisetype="transition"):
    """
    Function for loading the parameters from a DIDV class object.

    Parameters
    ----------
    DIDVobj : Object
        A DIDV class object after a fit has been run, such that there
        are Irwin parameters that can be used to model the noise.
    G : float, optional
        The thermal conductance of the TES in W/K
    qetbias : float, optional
        The QET bias in Amps
    tc : float
        The critical temperature of the TES in K
    tload : float
        The effective temperature of the load resistor in K
    tbath : float
        The bath temperature in K
    squiddc : float, optional
        The DC value of the SQUID and downstream electronics noise, in
        Amps/rtHz. The SQUID/electronics noise should have been fit
        beforehand, using the following model:
            (squiddc*(1.0+(squidpole/f)**squidn))**2.0
    squidpole : float, optional
        The frequency pole for the SQUID and downstream electronics
        noise, in Hz. The SQUID/electronics oise should have been fit
        beforehand, using the following model:
            (squiddc*(1.0+(squidpole/f)**squidn))**2.0
    squidn : float, optional
        The power of the SQUID and downstream electronics noise, in Hz.
        The SQUID/electronics noise should have been fit beforehand,
        using the following model:
            (squiddc*(1.0+(squidpole/f)**squidn))**2.0
    rnormal : float, optional
        The normal resistance of the TES in Ohms, only used if
        `noisetype` is 'normal'. Must be passed explicitly, as the dIdV
        fitting code does not fit it.
    noisetype : str, optional
        The type of the noise that is to be loaded. The options are
        transition : Use the Irwin parameters from the two pole fit as
            the transition noise model
        superconducting : Use the Irwin parameters from the one pole
            fit as the superconducting noise model
        normal : Use the Irwin parameters from the one pole fit as the
            normal noise model

    Returns
    -------
    TESobj : Object
        A TESnoise class object with all of the fit parameters loaded.

    """

    if noisetype == "superconducting":
        fitresult = DIDVobj.fitresult(1)
        if 'smallsignalparams' in fitresult:
            key = 'smallsignalparams'
        else:
            key = 'params'
        didv_dict = fitresult[key]
        rshunt = didv_dict['rsh']
        rload = didv_dict['rsh'] + didv_dict['rp']
        inductance = didv_dict['L']
        r0 = 0
        beta = 0
        loopgain = 0
        tau0 = 0
        G = 0
    elif noisetype == "normal":
        raise ValueError('Please specify rnormal.')
        fitresult = DIDVobj.fitresult(1)
        if 'smallsignalparams' in fitresult:
            key = 'smallsignalparams'
        else:
            key = 'params'
        didv_dict = fitresult[key]
        rshunt = didv_dict['rsh']
        rload = didv_dict['rsh'] + didv_dict['rp'] - rnormal
        inductance = didv_dict['L']
        r0 = rnormal
        beta = 0
        loopgain = 0
        tau0 = 0
        G = 0
    elif noisetype == "transition":
        fitresult = DIDVobj.fitresult(2)
        if 'smallsignalparams' in fitresult:
            key = 'smallsignalparams'
        else:
            key = 'params'
        didv_dict = fitresult[key]
        rshunt = didv_dict['rsh']
        rload = didv_dict['rsh'] + didv_dict['rp']
        inductance = didv_dict['L']
        r0 = didv_dict['r0']
        beta = didv_dict['beta']
        loopgain = didv_dict['l']
        tau0 = didv_dict['tau0']
    else:
        raise ValueError("Unrecognized noisetype")

    TESobj = TESnoise(
        rload=rload,
        r0=r0,
        rshunt=rshunt,
        inductance=inductance,
        beta=beta,
        loopgain=loopgain,
        tau0=tau0,
        G=G,
        qetbias=qetbias,
        tc=tc,
        tload=tload,
        tbath=tbath,
        squiddc=squiddc,
        squidpole=squidpole,
        squidn=squidn,
    )

    return TESobj


class TESnoise:
    """
    Class for the simulation of the TES noise using the simple Irwin theory. Supports noise simulation for 
    in transition, superconducting, and normal.
    
    Attributes
    ----------
    freqs : float, array_like
        The frequencies for which we will calculate the noise simulation
    rload : float
        The load resistance of the TES (sum of shunt and parasitic resistances) in Ohms
    r0 : float
        The bias resistance of the TES in Ohms
    rshunt : float
        The shunt resistance of the TES circuit in Ohms
    beta : float
        The current sensitivity of the TES (dlogR/dlogI), unitless
    loopgain : float
        The Irwin loop gain of the TES, unitless
    inductance : float
        The inductance of the TES circuit in Henries
    tau0 : float
        The thermal time constant (equals C/G) in s
    G : float
        The thermal conductance of the TES in W/K
    qetbias : float
        The QET bias in Amps
    tc : float
        The critical temperature of the TES in K
    tload : float
        The effective temperature of the load resistor in K
    tbath : float
        The bath temperature in K
    n : float
        The power-law dependence of the power flow to the heat bath
    lgcb : boolean
        Boolean flag that determines whether we use the ballistic (True) or
        diffusive limit when calculating TFN power noise
    squiddc : float
        The frequency pole for the SQUID and downstream electronics noise, in Hz. The SQUID/electronics
        noise should have been fit beforehand, using the following model:
            (squiddc*(1.0+(squidpole/f)**squidn))**2.0
    squidpole : float
        The frequency pole for the SQUID and downstream electronics noise, in Hz. The SQUID/electronics
        noise should have been fit beforehand, using the following model:
            (squiddc*(1.0+(squidpole/f)**squidn))**2.0
    squidn : float
        The power of the SQUID and downstream electronics noise, in Hz. The SQUID/electronics
        noise should have been fit beforehand, using the following model:
            (squiddc*(1.0+(squidpole/f)**squidn))**2.0
    f_tfn : float
        Function that estimates the noise suppression of the thermal fluctuation noise due
        to the difference in temperature between the bath and the TES. Supports the ballistic
        and diffusive limits, which is chosen via lgcb
        
    """

    def __init__(self, freqs=None, rload=0.012, r0=0.150, rshunt=0.005, beta=1.0, loopgain=10.0, 
                 inductance=400.0e-9, tau0=500.0e-6, G=5.0e-10, qetbias=160e-6, tc=0.040, tload=0.9, 
                 tbath=0.020, n=5.0, lgcb=True, squiddc=2.5e-12, squidpole=0.0, squidn=1.0):
        """
        Initialization of the TES noise class.

        Parameters
        ----------
        freqs : float, ndarray, optional
            The frequencies for which we will calculate the noise simulation
        rload : float, optional
            The load resistance of the TES (sum of shunt and parasitic resistances) in Ohms
        r0 : float, optional
            The bias resistance of the TES in Ohms
        rshunt : float, optional
            The shunt resistance of the TES circuit in Ohms
        beta : float, optional
            The current sensitivity of the TES (dlogR/dlogI), unitless
        loopgain : float, optional
            The Irwin loop gain of the TES, unitless
        inductance : float, optional
            The inductance of the TES circuit in Henries
        tau0 : float, optional
            The thermal time constant (equals C/G) in s
        G : float, optional
            The thermal conductance of the TES in W/K
        qetbias : float, optional
            The QET bias in Amps
        tc : float, optional
            The critical temperature of the TES in K
        tload : float, optional
            The effective temperature of the load resistor in K
        tbath : float, optional
            The bath temperature in K
        n : float, optional
            The power-law dependence of the power flow to the heat bath
        lgcb : boolean, optional
            Boolean flag that determines whether we use the ballistic (True) or
            diffusive limit when calculating TFN power noise
        squiddc : float, optional
            The frequency pole for the SQUID and downstream electronics noise, in Hz. The SQUID/electronics
            noise should have been fit beforehand, using the following model:
                (squiddc*(1.0+(squidpole/f)**squidn))**2.0
        squidpole : float, optional
            The frequency pole for the SQUID and downstream electronics noise, in Hz. The SQUID/electronics
            noise should have been fit beforehand, using the following model:
                (squiddc*(1.0+(squidpole/f)**squidn))**2.0
        squidn : float, optional
            The power of the SQUID and downstream electronics noise, in Hz. The SQUID/electronics
            noise should have been fit beforehand, using the following model:
                (squiddc*(1.0+(squidpole/f)**squidn))**2.0

        """
    
        if freqs is None:
            self.freqs = np.logspace(0, 5.5, 10000)
        else:
            self.freqs = freqs
        self.rload = rload
        self.r0 = r0
        self.rshunt = rshunt
        self.beta = beta
        self.loopgain = loopgain
        self.inductance = inductance
        self.tau0 = tau0
        self.qetbias = qetbias
        self.i0 = self.qetbias*self.rshunt/(self.r0+self.rload)
        self.tc = tc
        self.tload = tload
        self.G = G
        self.tbath = tbath
        self.n = n
        self.lgcb = lgcb
        self.squiddc = squiddc
        self.squidpole = squidpole
        self.squidn = squidn
        
        if self.lgcb: # ballistic limit
            self.f_tfn = ((self.tbath/self.tc)**(self.n+1.0)+1.0)/2.0
        else:         # diffusive limit
            self.f_tfn = self.n/(2.0*self.n+1.0) * ((self.tbath/self.tc)**(2.0*self.n+1.0)-1.0)/((self.tbath/self.tc)**(self.n)-1.0)

    def dIdV(self, freqs=None):
        """
        The two-pole dIdV function determined from the TES parameters.
        
        Parameters
        ----------
        freqs : float, ndarray, optional
            The frequencies for which we will calculate the noise simulation. If left as None, the 
            function will use the values from the initialization.
                
        Returns
        -------
        dIdV : float, ndarray
            The two-pole dIdV function
                
        """
        
        if freqs is None:
            freqs = self.freqs
        omega = 2.0*np.pi*freqs
        dVdI = self.rload+self.r0*(1.0+self.beta)+1.0j*omega*self.inductance+self.r0*self.loopgain/(1.0-self.loopgain)*(2.0+self.beta)/(1.0+1.0j*omega*self.tau0/(1.0-self.loopgain))
        return 1.0/dVdI

    def dIdP(self, freqs=None):
        """
        The two-pole dIdP function determined from the TES parameters.
        
        Parameters
        ----------
        freqs : float, ndarray, optional
            The frequencies for which we will calculate the noise simulation. If left as None, the 
            function will use the values from the initialization.
                
        Returns
        -------
        dIdP : float, ndarray
            The two-pole dIdP function
                
        """
        
        if freqs is None:
            freqs = self.freqs
        omega = 2.0*np.pi*freqs
        # return -self.G*self.loopgain/(self.i0*self.inductance*self.tau0*self.G) / ((1.0/(self.inductance/(self.rload+(1.0+self.beta)*self.r0))+1.0j*omega)*(1.0/(self.tau0/(1.0-self.loopgain))+1.0j*omega)+self.loopgain*self.G/(self.inductance*self.tau0*self.G) * self.r0*(2.0+self.beta))
        return 1.0/self.i0 * 1.0/(1.0-1.0/self.loopgain) * 1.0/(1.0+1.0j*omega*self.tau0/(1.0-self.loopgain))*self.dIdV(freqs)

    def s_vload(self, freqs=None):
        """
        The Johnson load voltage noise determined from the TES parameters. This formula holds no
        matter where we are in transition.
        
        Parameters
        ----------
        freqs : float, ndarray, optional
            The frequencies for which we will calculate the noise simulation. If left as None, the 
            function will use the values from the initialization.
                
        Returns
        -------
        s_vload : float, ndarray
            The Johnson load voltage noise at the specified frequencies
                
        """
        
        if freqs is None:
            freqs = self.freqs
        return 4.0*constants.k*self.tload*self.rload * np.ones_like(freqs)

    def s_iload(self, freqs=None):
        """
        The Johnson load current noise determined from the TES parameters for in transition.
        
        Parameters
        ----------
        freqs : float, ndarray, optional
            The frequencies for which we will calculate the noise simulation. If left as None, the 
            function will use the values from the initialization.
                
        Returns
        -------
        s_iload : float, ndarray
            The Johnson load current noise at the specified frequencies
             
             
        """
        
        if freqs is None:
            freqs = self.freqs
        return self.s_vload(freqs)*np.abs(self.dIdV(freqs))**2.0

    def s_pload(self, freqs=None):
        """
        The Johnson load power noise determined from the TES parameters for in transition.
        
        Parameters
        ----------
        freqs : float, ndarray, optional
            The frequencies for which we will calculate the noise simulation. If left as None, the 
            function will use the values from the initialization.
                
        Returns
        -------
        s_pload : float, ndarray
            The Johnson load power noise at the specified frequencies
                
        """
        
        if freqs is None:
            freqs = self.freqs
        return self.s_iload(freqs)/np.abs(self.dIdP(freqs))**2.0

    def s_vtes(self, freqs=None):
        """
        The Johnson TES voltage noise determined from the TES parameters for in transition.
        
        Parameters
        ----------
        freqs : float, ndarray, optional
            The frequencies for which we will calculate the noise simulation. If left as None, the 
            function will use the values from the initialization.
                
        Returns
        -------
        s_vtes : float, ndarray
            The Johnson TES voltage noise at the specified frequencies
                
        """
        
        if freqs is None:
            freqs = self.freqs
        return 4.0*constants.k*self.tc*self.r0*(1.0+self.beta)**2.0 * np.ones_like(freqs)

    def s_ites(self, freqs=None):
        """
        The Johnson TES current noise determined from the TES parameters for in transition. 
        This noise has both an electronic and thermal component.
        
        Parameters
        ----------
        freqs : float, ndarray, optional
            The frequencies for which we will calculate the noise simulation. If left as None, the 
            function will use the values from the initialization.
                
        Returns
        -------
        s_ites : float, ndarray
            The Johnson TES current noise at the specified frequencies
                
        """
        
        if freqs is None:
            freqs = self.freqs
        return self.s_vtes(freqs)*np.abs(self.dIdV(freqs)-self.i0*self.dIdP(freqs))**2.0

    def s_ptes(self, freqs=None):
        """
        The Johnson TES power noise determined from the TES parameters for in transition.
        
        Parameters
        ----------
        freqs : float, ndarray, optional
            The frequencies for which we will calculate the noise simulation. If left as None, the 
            function will use the values from the initialization.
                
        Returns
        -------
        s_ptes : float, ndarray
            The Johnson TES power noise at the specified frequencies
                
        """
        
        if freqs is None:
            freqs = self.freqs
        return self.s_ites(freqs)/np.abs(self.dIdP(freqs))**2.0

    def s_ptfn(self, freqs=None):
        """
        The thermal fluctuation noise in power determined from the TES parameters for in transition.
        
        Parameters
        ----------
        freqs : float, ndarray, optional
            The frequencies for which we will calculate the noise simulation. If left as None, the 
            function will use the values from the initialization.
                
        Returns
        -------
        s_ptfn : float, ndarray
            The thermal fluctuation noise in power at the specified frequencies
                
        """
        
        if freqs is None:
            freqs = self.freqs
        return 4.0*constants.k*self.tc**2.0 * self.G * self.f_tfn * np.ones_like(freqs)

    def s_itfn(self, freqs=None):
        """
        The thermal fluctuation noise in current determined from the TES parameters for in transition.
        
        Parameters
        ----------
        freqs : float, ndarray, optional
            The frequencies for which we will calculate the noise simulation. If left as None, the 
            function will use the values from the initialization.
                
        Returns
        -------
        s_itfn : float, ndarray
            The thermal fluctuation noise in current at the specified frequencies
                
        """
        
        if freqs is None:
            freqs = self.freqs
        return self.s_ptfn(freqs)*np.abs(self.dIdP(freqs))**2.0

    def s_isquid(self, freqs=None):
        """
        The SQUID and downstream electronics current noise, currently is using a 1/f model that
        must be specified when initializing the class.
        
        Parameters
        ----------
        freqs : float, ndarray, optional
            The frequencies for which we will calculate the noise simulation. If left as None, the 
            function will use the values from the initialization.
                
        Returns
        -------
        s_isquid : float, ndarray
            The SQUID and downstream electronics current noise at the specified frequencies
                
        """
        
        if freqs is None:
            freqs = self.freqs
        return (self.squiddc*(1.0+(self.squidpole/freqs)**self.squidn))**2.0

    def s_psquid(self, freqs=None):
        """
        The SQUID and downstream electronics power noise, currently is using a 1/f model that
        must be specified when initializing the class. This is only used for when the TES is
        in transition.
        
        Parameters
        ----------
        freqs : float, ndarray, optional
            The frequencies for which we will calculate the noise simulation. If left as None, the 
            function will use the values from the initialization.
                
        Returns
        -------
        s_psquid : float, ndarray
            The SQUID and downstream electronics power noise at the specified frequencies
                
        """
        
        if freqs is None:
            freqs = self.freqs
        return self.s_isquid(freqs)/np.abs(self.dIdP(freqs))**2.0

    def s_itot(self, freqs=None):
        """
        The total current noise for the TES in transition. This is calculated by summing each of
        the current noise sources together. Units are [A^2/Hz].
        
        Parameters
        ----------
        freqs : float, ndarray, optional
            The frequencies for which we will calculate the noise simulation. If left as None, the 
            function will use the values from the initialization.
                
        Returns
        -------
        s_itot : float, ndarray
            The total current noise at the specified frequencies
                
        """
        
        if freqs is None:
            freqs = self.freqs
        return self.s_iload(freqs)+self.s_ites(freqs)+self.s_itfn(freqs)+self.s_isquid(freqs)

    def s_ptot(self, freqs=None): # total power noise [W^2/Hz]
        """
        The total power noise for the TES in transition. This is calculated by summing each of
        the current noise sources together and using dIdP to convert to power noise. Units are [W^2/Hz].
        
        Parameters
        ----------
        freqs : float, ndarray, optional
            The frequencies for which we will calculate the noise simulation. If left as None, the 
            function will use the values from the initialization.
                
        Returns
        -------
        s_ptot : float, ndarray
            The total power noise at the specified frequencies
                
        """
        
        if freqs is None:
            freqs = self.freqs
        return self.s_itot(freqs)/np.abs(self.dIdP(freqs))**2.0
    
    def dIdVnormal(self, freqs=None):
        """
        The one-pole dIdV function determined from the TES parameters for when the TES
        is normal.
        
        Parameters
        ----------
        freqs : float, ndarray, optional
            The frequencies for which we will calculate the noise simulation. If left as None, the 
            function will use the values from the initialization.
                
        Returns
        -------
        dIdVnormal : float, ndarray
            The one-pole dIdV function for when the TES is normal.
                
        """
        
        if freqs is None:
            freqs = self.freqs
        omega = 2.0*np.pi*freqs
        return 1.0/(self.rload+self.r0+1.0j*omega*self.inductance)
    
    def s_iloadnormal(self, freqs=None):
        """
        The Johnson load current noise determined from the TES parameters for normal.
        
        Parameters
        ----------
        freqs : float, ndarray, optional
            The frequencies for which we will calculate the noise simulation. If left as None, the 
            function will use the values from the initialization.
                
        Returns
        -------
        s_iloadnormal : float, ndarray
            The Johnson load current noise at the specified frequencies
                
        """
        
        if freqs is None:
            freqs = self.freqs
        return self.s_vload(freqs)*np.abs(self.dIdVnormal(freqs))**2.0
    
    def s_vtesnormal(self, freqs=None):
        """
        The Johnson TES voltage noise determined from the TES parameters for normal.
        
        Parameters
        ----------
        freqs : float, ndarray, optional
            The frequencies for which we will calculate the noise simulation. If left as None, the 
            function will use the values from the initialization.
                
        Returns
        -------
        s_vtesnormal : float, ndarray
            The Johnson TES voltage noise at the specified frequencies
                
        """
        
        if freqs is None:
            freqs = self.freqs
        return 4.0*constants.k*self.tc*self.r0 * np.ones_like(freqs)

    def s_itesnormal(self, freqs=None):
        """
        The Johnson TES current noise determined from the TES parameters for normal.
        
        Parameters
        ----------
        freqs : float, ndarray, optional
            The frequencies for which we will calculate the noise simulation. If left as None, the 
            function will use the values from the initialization.
                
        Returns
        -------
        s_itesnormal : float, ndarray
            The Johnson TES current noise at the specified frequencies
                
        """
        
        if freqs is None:
            freqs = self.freqs
        return self.s_vtesnormal(freqs)*np.abs(self.dIdVnormal(freqs))**2.0
    
    def s_itotnormal(self, freqs=None):
        """
        The total current noise for the TES when normal. This is calculated by summing each of
        the current noise sources together. Units are [A^2/Hz].
        
        Parameters
        ----------
        freqs : float, ndarray, optional
            The frequencies for which we will calculate the noise simulation. If left as None, the 
            function will use the values from the initialization.
                
        Returns
        -------
        s_itotnormal : float, ndarray
            The total current noise at the specified frequencies
                
        """
        
        if freqs is None:
            freqs = self.freqs
        return self.s_iloadnormal(freqs)+self.s_itesnormal(freqs)+self.s_isquid(freqs)
    
    def dIdVsc(self, freqs=None):
        """
        The one-pole dIdV function determined from the TES parameters for when the TES
        is superconducting.
        
        Parameters
        ----------
        freqs : float, ndarray, optional
            The frequencies for which we will calculate the noise simulation. If left as None, the 
            function will use the values from the initialization.
                
        Returns
        -------
        dIdVsc : float, ndarray
            The one-pole dIdV function for when the TES is superconducting.
                
        """
        
        if freqs is None:
            freqs = self.freqs
        omega = 2.0*np.pi*freqs
        return 1.0/(self.rload+1.0j*omega*self.inductance)
    
    def s_iloadsc(self, freqs=None):
        """
        The Johnson load current noise determined from the TES parameters for superconducting.
        
        Parameters
        ----------
        freqs : float, ndarray, optional
            The frequencies for which we will calculate the noise simulation. If left as None, the 
            function will use the values from the initialization.
                
        Returns
        -------
        s_iloadsc : float, ndarray
            The Johnson load current noise at the specified frequencies
                
        """
        
        if freqs is None:
            freqs = self.freqs
        return self.s_vload(freqs)*np.abs(self.dIdVsc(freqs))**2.0
    
    def s_itotsc(self, freqs=None):
        """
        The total current noise for the TES when superconducting. This is calculated by summing each of
        the current noise sources together. Units are [A^2/Hz].
        
        Parameters
        ----------
        freqs : float, ndarray, optional
            The frequencies for which we will calculate the noise simulation. If left as None, the 
            function will use the values from the initialization.
                
        Returns
        -------
        s_itotsc : float, ndarray
            The total current noise at the specified frequencies
                
        """
        
        if freqs is None:
            freqs = self.freqs
        return self.s_iloadsc(freqs)+self.s_isquid(freqs)




import numpy as np
import matplotlib.pyplot as plt


__all__ = ["_plot_iv", "_plot_rv", "_plot_pv", "_plot_all_curves"]


def _plot_iv(IVobject, temps="all", chans="all", showfit=True, lgcsave=False, savepath="", savename=""):
    """
    Function to plot the IV curves for the data in an IV (IBIS) object.
    
    Parameters
    ----------
    IVobject : class
        The IV class object that the data is stored in.
    temps : string, array_like, int, optional
        Which bath temperatures to plot. Setting to "all" plots all of them. Can also set
        to a subset of bath temperatures, or just one
    chans : string, array_like, int, optional
        Which bath temperatures to plot. Setting to "all" plots all of them. Can also set
        to a subset of bath temperatures, or just one
    showfit : boolean, optional
        Boolean flag to also plot the linear fit to the normal data
    lgcsave : boolean, optional
        Boolean flag to save the plot
    savepath : string, optional
        Path to save the plot to, saves it to the current directory by default
    savename : string, optional
        Name to append to the plot file name, if saving

    """
    
    ntemps, nch, niters = IVobject.dites.shape
    chan_names = IVobject.chan_names
    
    if temps == "all":
        trange = range(ntemps)
    elif np.isscalar(temps):
        trange = np.array([temps])
    else:
        trange = np.array(temps)
        
    if chans == "all":
        chrange = range(nch)
    elif np.isscalar(chans):
        chrange = np.array([chans])
    else:
        chrange = np.array(chans)
    
    ch_colors = plt.cm.plasma(np.linspace(0, 1, num=len(trange)*len(chrange)+1))[:-1]
    
    fig, ax = plt.subplots(figsize=(6,4))
    
    for it, t in enumerate(trange):
        for ich, ch in enumerate(chrange):
            if ntemps > 1:
                label_str="Temp {}, Channel {}".format(t,chan_names[ch])
            else:
                label_str="Channel {}".format(chan_names[ch])
            ax.scatter(IVobject.vb[t, ch]*1e6, IVobject.ites[t, ch]*1e6,label=label_str ,
                        color=ch_colors[it*len(chrange) + ich], s=20.0)
            ax.plot(IVobject.vb[t, ch]*1e6, IVobject.ites[t, ch]*1e6, color=ch_colors[it*len(chrange) + ich],
                    linewidth=2.5, alpha=0.5)
            ax.errorbar(IVobject.vb[t, ch]*1e6, IVobject.ites[t, ch]*1e6, yerr=IVobject.ites_err[t, ch]*1e6,
                        xerr=IVobject.vb_err[t, ch]*1e6,
                         linestyle='None', color='k')
            if showfit:
                maxind = np.argmax(abs(IVobject.vb[t, ch]))
                
                if IVobject.vb[t, ch, maxind] > 0:
                    vbfit = np.linspace(0, max(IVobject.vb[t,ch]*1e6), num=10)
                else:
                    vbfit = np.linspace(min(IVobject.vb[t,ch]*1e6), 0, num=10)
                    
                ax.plot(vbfit, 1.0/(IVobject.rfit[t, ch])*vbfit, color=ch_colors[it*len(chrange) + ich], alpha=0.1)

    ax.legend(loc='best')
    ax.set_xlabel(r'Bias Voltage [$\mu V$]', fontsize=14)
    ax.set_ylabel(r'Current through TES [$\mu A$]', fontsize=14)
    ax.grid(linestyle='dotted')
    ax.tick_params(which='both',direction='in',right=True,top=True)
    ax.set_title("$I_0$ vs. $V_b$", fontsize=16)
    
    if lgcsave:
        fig.savefig(savepath+"iv_curve_{}.png".format(savename))
        plt.close(fig)
    else:
        #plt.show()
        return fig, ax

def _plot_rv(IVobject, temps="all", chans="all", lgcsave=False, savepath="", savename=""):
    """
    Function to plot the resistance curves for the data in an IV object.
    
    Parameters
    ----------
    IVobject : class
        The IV class object that the data is stored in.
    temps : string, array_like, int, optional
        Which bath temperatures to plot. Setting to "all" plots all of them. Can also set
        to a subset of bath temperatures, or just one
    chans : string, array_like, int, optional
        Which bath temperatures to plot. Setting to "all" plots all of them. Can also set
        to a subset of bath temperatures, or just one
    lgcsave : boolean, optional
        Boolean flag to save the plot
    savepath : string, optional
        Path to save the plot to, saves it to the current directory by default
    savename : string, optional
        Name to append to the plot file name, if saving

    """

    ntemps, nch, niters = IVobject.r0.shape
    chan_names = IVobject.chan_names
    
    if temps == "all":
        trange = range(ntemps)
    elif np.isscalar(temps):
        trange = np.array([temps])
    else:
        trange = np.array(temps)
        
    if chans == "all":
        chrange = range(nch)
    elif np.isscalar(chans):
        chrange = np.array([chans])
    else:
        chrange = np.array(chans)
    
    ch_colors = plt.cm.plasma(np.linspace(0, 1, num=len(trange)*len(chrange)+1))[:-1]
    
    fig, ax = plt.subplots(figsize=(6,4))
    
    for it, t in enumerate(trange):
        for ich, ch in enumerate(chrange):
            if ntemps > 1:
                label_str="Temp {}, Channel {}".format(t,chan_names[ch])
            else:
                label_str="Channel {}".format(chan_names[ch])
            ax.scatter(IVobject.vb[t, ch]*1e6, IVobject.r0[t, ch]*1e3, label=label_str,
                        color=ch_colors[it*len(chrange) + ich], s=20.0)
            ax.plot(IVobject.vb[t, ch]*1e6, IVobject.r0[t, ch]*1e3, color=ch_colors[it*len(chrange) + ich], 
                    linewidth=2.5, alpha=0.5)
            ax.errorbar(IVobject.vb[t, ch]*1e6, IVobject.r0[t, ch]*1e3, yerr=IVobject.r0_err[t, ch]*1e3,
                        xerr=IVobject.vb_err[t, ch]*1e6,
                         linestyle='None', color='k')

    ax.legend(loc='best')
    ax.set_xlabel(r'Bias Voltage [$\mu V$]', fontsize=14)
    ax.set_ylabel(r'Resistance of TES [$m \Omega$]', fontsize=14)
    ax.grid(linestyle='dotted')
    ax.tick_params(which='both',direction='in',right=True,top=True)
    ax.set_title(r"$R_0$ vs. $V_b$", fontsize=16)
    
    if lgcsave:
        fig.savefig(savepath+"rv_curve_{}.png".format(savename))
        plt.close(fig)
    else:
        #plt.show()
        return fig, ax    

def _plot_pv(IVobject,  temps="all", chans="all", lgcsave=False, savepath="", savename=""):
    """
    Function to plot the power curves for the data in an IV object.
    
    Parameters
    ----------
    IVobject : class
        The IV class object that the data is stored in.
    temps : string, array_like, int, optional
        Which bath temperatures to plot. Setting to "all" plots all of them. Can also set
        to a subset of bath temperatures, or just one
    chans : string, array_like, int, optional
        Which bath temperatures to plot. Setting to "all" plots all of them. Can also set
        to a subset of bath temperatures, or just one
    lgcsave : boolean, optional
        Boolean flag to save the plot
    savepath : string, optional
        Path to save the plot to, saves it to the current directory by default
    savename : string, optional
        Name to append to the plot file name, if saving

    """

    ntemps, nch, niters = IVobject.ptes.shape
    chan_names = IVobject.chan_names
    
    if temps == "all":
        trange = range(ntemps)
    elif np.isscalar(temps):
        trange = np.array([temps])
    else:
        trange = np.array(temps)
        
    if chans == "all":
        chrange = range(nch)
    elif np.isscalar(chans):
        chrange = np.array([chans])
    else:
        chrange = np.array(chans)
    
    ch_colors = plt.cm.plasma(np.linspace(0, 1, num=len(trange)*len(chrange)+1))[:-1]
    
    fig, ax = plt.subplots(figsize=(6,4))
    
    for it, t in enumerate(trange):
        for ich, ch in enumerate(chrange):
            if ntemps > 1:
                label_str="Temp {}, Channel {}".format(t,chan_names[ch])
            else:
                label_str="Channel {}".format(chan_names[ch])
            ax.scatter(IVobject.vb[t, ch]*1e6, IVobject.ptes[t, ch]*1e12, label=label_str,
                        color=ch_colors[it*len(chrange) + ich], s=20.0)
            ax.plot(IVobject.vb[t, ch]*1e6, IVobject.ptes[t, ch]*1e12, color=ch_colors[it*len(chrange) + ich], 
                    linewidth=2.5, alpha=0.5)
            ax.errorbar(IVobject.vb[t, ch]*1e6, IVobject.ptes[t, ch]*1e12, yerr=IVobject.ptes_err[t, ch]*1e12,
                        xerr=IVobject.vb_err[t, ch]*1e6,
                         linestyle='None', color='k')

    ax.legend(loc='best')
    ax.set_xlabel(r'Bias Voltage [$\mu V$]', fontsize=14)
    ax.set_ylabel(r'Power [$pW$]', fontsize=14)
    ax.grid(linestyle='dotted')
    ax.tick_params(which='both',direction='in',right=True,top=True)
    ax.set_title("$P_0$ vs. $V_b$", fontsize=16)
    
    if lgcsave:
        fig.savefig(savepath+"pv_curve_{}.png".format(savename))
        plt.close(fig)
    else:
        #plt.show()
        return fig, ax
        
def _plot_all_curves(IVobject,  temps="all", chans="all", showfit=True, lgcsave=False, savepath="", savename=""):
    """
    Function to plot the IV, resistance, and power curves for the data in an IV object.
    
    Parameters
    ----------
    IVobject : class
        The IV class object that the data is stored in.
    temps : string, array_like, int, optional
        Which bath temperatures to plot. Setting to "all" plots all of them. Can also set
        to a subset of bath temperatures, or just one
    chans : string, array_like, int, optional
        Which bath temperatures to plot. Setting to "all" plots all of them. Can also set
        to a subset of bath temperatures, or just one
    showfit : boolean, optional
        Boolean flag to also plot the linear fit to the normal data
    lgcsave : boolean, optional
        Boolean flag to save the plot
    savepath : string, optional
        Path to save the plot to, saves it to the current directory by default
    savename : string, optional
        Name to append to the plot file name, if saving

    """
    
    _plot_iv(IVobject,  temps=temps, chans=chans, showfit=showfit, lgcsave=lgcsave, savepath=savepath, savename=savename)
    _plot_rv(IVobject,  temps=temps, chans=chans, lgcsave=lgcsave, savepath=savepath, savename=savename)
    _plot_pv(IVobject,  temps=temps, chans=chans, lgcsave=lgcsave, savepath=savepath, savename=savename)


  